# -*- coding: utf-8 -*-
"""Finite languages and related automata manipulation

Finite languages manipulation

.. *Authors:* Rogério Reis & Nelma Moreira

.. *This is part of FAdo project*   <http://fado.dcc.fc.up.pt

.. *Version:* 1.3.3

.. *Copyright*: 1999-2014 Rogério Reis & Nelma Moreira {rvr,nam}@dcc.fc.up.pt

.. This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA."""

import fa
from copy import copy
from common import *
import random


class FL(object):
    """Finite Language Class

    :var Words: the elements of the language
    :var Sigma: the alphabet"""
    def __init__(self, wordsList=None, Sigma=None):
        self.Words = set([])
        if Sigma is None:
            self.Sigma = set([])
        else:
            self.Sigma = Sigma
        if wordsList is not None:
            for w in wordsList:
                if type(w) is Word:
                    self.addWord(w)
                else:
                    self.addWord(Word(w))

    def __str__(self):
        l = "({"
        for i in self.Words:
            l += str(i) + ","
        l = l[:-1] + "}, {"
        for i in self.Sigma:
            l += str(i)+","
        l = l[:-1] + "})"
        return l

    def __repr__(self):
        return "FL%s" % self.__str__()

    def __len__(self):
        return len(self.Words)

    def __contains__(self, item):
        return item in self.Words

    def union(self, other):
        """union of FL:   a | b

        :param FL other: right hand operand
        :rtype: FL
        :raises FAdoGeneralError: if both arguments are not FL"""
        return self.__or__(other)

    def __or__(self, other):
        if type(other) != type(self):
            raise FAdoGeneralError("Incompatible objects")
        new = FL()
        new.Sigma = self.Sigma | other.Sigma
        new.Words = self.Words | other.Words
        return new

    def intersection(self, other):
        """Intersection of FL: a & b

        :param FL other: right hand operand
        :raises FAdoGeneralError: if both arguments are not FL"""
        return self.__and__(other)

    def __iter__(self):
        return iter(self.Words)

    def __and__(self, other):
        if type(other) != type(self):
            raise FAdoGeneralError("Incompatible objects")
        new = FL()
        new.Sigma = self.Sigma | other.Sigma
        new.Words = self.Words & other.Words
        return new

    def diff(self, other):
        """Difference of FL: a - b

        :param FL other: right hand operand
        :rtype: FL
        :raises FAdoGeneralError: if both arguments are not FL"""
        return self.__sub__(other)

    def __sub__(self, other):
        if type(other) != type(self):
            raise FAdoGeneralError("Incompatible objects")
        new = FL()
        new.Sigma = self.Sigma | other.Sigma
        new.Words = self.Words - other.Words
        return new

    def setSigma(self, Sigma, Strict=False):
        """Sets the alphabet of a FL

        :param set Sigma: alphabet
        :param bool Strict: behaviour

        .. attention::
           Unless Strict flag is set to True, alphabet can only be enlarged.  The resulting alphabet is  in fact the
           union of the former alphabet with the new one. If flag is set to True, the alphabet is simply replaced."""
        if Strict:
            self.Sigma = Sigma
        else:
            self.Sigma = self.Sigma.union(Sigma)

    def addWords(self, wList):
        """Adds a list of words to a FL

        :param list wList: words to add"""
        self.Words |= set(wList)
        for w in wList:
            if w.epsilonP():
                continue
            for c in w:
                self.Sigma.add(c)

    def addWord(self, word):
        """Adds a word to a FL
        :type word: Word
        :rtype: FL"""
        if word in self:
            return
        self.Words.add(word)
        if not word.epsilonP():
            for c in word:
                self.Sigma.add(c)

    def suffixClosedP(self):
        """Tests if a language is suffix closed

        :rtype: bool"""
        wrds = list(copy(self.Words))
        if Epsilon not in wrds:
            return False
        else:
            wrds.remove(Epsilon)
        wrds.sort(lambda x, y: len(x) - len(y))
        while wrds:
            w = wrds.pop()
            for w1 in suffixes(w):
                if w1 not in self.Words:
                    return False
                else:
                    if w1 in wrds:
                        wrds.remove(w1)
        return True

    def filter(self, automata):
        """Separates a language in two other using a DFA of NFA as a filter

        :param DFA|NFA automata: the automata to be used as a filter
        :returns: the accepted/unaccepted pair of languages
        :rtype: tuple of FL"""
        a, b = (FL(), FL())
        a.setSigma(self.Sigma)
        b.setSigma(self.Sigma)
        for w in self.Words:
            if automata.evalWord(w):
                a.addWords([w])
            else:
                b.addWords([w])
        return a, b

    def MADFA(self):
        """Generates the minimal acyclical DFA using specialized algorithm

        .. versionadded:: 1.3.3

        .. seealso:: Incremental Construction of Minimal Acyclic Finite-State Automata, J.Daciuk, S.Mihov, B.Watson and R.E.Watson

        :rtype: ADFA"""
        aut = ADFA()
        i = aut.addState()
        aut.setInitial(i)
        aut.setSigma(self.Sigma)
        register = set()
        foo = sorted(list(self.Words))
        for w in foo:  # sorted(list(self.Words)):
            (cPrefix, lState) = aut._common_prefix(w)
            cSuffix = w[len(cPrefix):]
            if aut.delta.get(lState, {}):
                aut._replace_or_register(lState, register)
            aut.addSuffix(lState, cSuffix)
        aut._replace_or_register(i, register)
        aut.Minimal = True
        return aut

    def trieFA(self):
        """Generates the trie automaton that recognises this language

        :returns: the trie automaton
        :rtype: ADFA"""
        new = ADFA()
        new.setSigma(copy(self.Sigma))
        i = new.addState()
        new.setInitial(i)
        for w in self.Words:
            if w.epsilonP():
                new.addFinal(i)
            else:
                s = i
                for c in w:
                    if c not in new.delta.get(s, []):
                        sn = new.addState()
                        new.addTransition(s, c, sn)
                        s = sn
                    else:
                        s = new.delta[s][c]
                new.addFinal(s)
        return new

    def toDFA(self):
        """ Generates a DFA recognizing the language

        :rtype: ADFA

        .. versionadded:: 1.2"""
        return self.trieFA()

    def toNFA(self):
        """Generates a NFA recognizing the language

        :rtype: ANFA

        .. versionadded:: 1.2"""
        return self.toDFA().toANFA()

    # noinspection PyUnboundLocalVariable
    def multiLineAutomaton(self):
        """Generates the trivial linear ANFA equivalent to this language
    
        :rtype: ANFA"""
        new = ANFA()
        s1 = None
        new.setSigma(copy(self.Sigma))
        for w in self.Words:
            s = new.addState()
            new.addInitial(s)
            for c in w:
                s1 = new.addState()
                new.addTransition(s, c, s1)
                s = s1
            new.addFinal(s1)
        return new


class DFCA(fa.DFA):
    """Deterministic Cover Automata class

    .. inheritance-diagram:: DFCA"""

    def __init__(self):
        super(DFCA, self).__init__()
        self.length = None

    @property
    def length(self):
        """
        :return: size of the longest word
        :rtype: int"""
        return self.length

    @length.setter
    def length(self, value):
        """Setter
        :param int value: size"""
        self.length = value

    @length.deleter
    def length(self):
        """Length deleter"""
        self.length = None


class AFA(object):
    """Base class for Acyclic Finite Automata

    .. inheritance-diagram:: AFA

    .. note::
       This is just a container for some common methods. **Not to be used directly!!**"""
    def __init__(self):
        self.Dead = None
        self.delta = dict()
        self.Initial = None
        self.States = []
        self.Final = set()

    @abstractmethod
    def addState(self):
        """
        :rtype: int"""
        pass

    @abstractmethod
    def finalP(self):
        pass

    def setDeadState(self, sti):
        """Identifies the dead state

        :param int sti: index of the dead state

        .. attention::
           nothing is done to ensure that the state given is legitimate

        .. note::
           without dead state identified, most of the methods for acyclic automata can not be applied"""
        self.Dead = sti

    def ensureDead(self):
        """Ensures that a state is defined as dead"""
        try:
            _ = self.Dead
        except AttributeError:
            x = self.addState(DeadName)
            self.setDeadState(x)

    def ordered(self):
        """Orders states names in its topological order
    
        :returns: ordered list of state indexes
        :rtype: list of int

        .. note::
           one could use the FA.toposort() method, but special care must be taken with the dead state for the
           algorithms related with cover automata."""

        def _dealS(st1):
            if st1 not in torder:
                torder.append(st1)
                if st1 in self.delta.keys():
                    for k in self.delta[st1]:
                        for dest in forceIterable(self.delta[st1][k]):
                            if dest not in torder and dest != self.Dead:
                                queue.append(dest)

        try:
            dead = self.Dead
        except AttributeError:
            raise FAdoGeneralError("ADFA has not dead state identified")
        torder, queue = [], []
        _dealS(self.Initial)
        while queue:
            st = queue.pop()
            _dealS(st)
        torder.append(dead)
        return torder

    def _getRdelta(self):
        """
        :returns: pair, map of number of sons map, of reverse conectivity
        :rtype: dict"""
        done = set()
        deltaC, rdelta = {}, {}
        notDone = set(forceIterable(self.Initial))
        while notDone:
            sts = uSet(notDone)
            done.add(sts)
            l = set()
            for k in self.delta.get(sts, []):
                for std in forceIterable(self.delta[sts][k]):
                    l.add(std)
                    rdelta.setdefault(std, set([])).add(sts)
                    if std not in done:
                        notDone.add(std)
            deltaC[sts] = len(l)
            notDone.remove(sts)
        for s in forceIterable(self.Initial):
            if s not in rdelta:
                rdelta[s] = set()
        return deltaC, rdelta

    def directRank(self):
        """Compute rank function

        :return: ranf map
        :rtype: dict"""
        r, _ = self.evalRank()
        n = {}
        for x in r:
            for i in r[x]:
                n[i] = x
        return n

    def evalRank(self):
        """Evaluates the rank map of a automaton

        :return: pair of sets of states by rank map, reverse delta accessability map
        :rtype: tuple"""
        (deltaC, rdelta) = self._getRdelta()
        rank, deltai = {}, {}
        for s in xrange(len(self.States)):
            deltai.setdefault(deltaC[s], set([])).add(s)
        i = -1
        notDone = copy(range(len(self.States)))
        deltaC[self.Dead] = 0
        deltai[1].remove(self.Dead)
        deltai[0] = {self.Dead}
        rdelta[self.Dead].remove(self.Dead)
        while notDone:
            rank[i] = deepcopy(deltai[0])
            deltai[0] = set()
            for s in rank[i]:
                for s1 in rdelta[s]:
                    l = deltaC[s1]
                    deltaC[s1] = l - 1
                    deltai[l].remove(s1)
                    deltai.setdefault(l - 1, set()).add(s1)
                notDone.remove(s)
            i += 1
        return rank, rdelta

    def getLeaves(self):
        """The set of leaves, i.e. final states for last symbols of language words

        :return: set of leaves
        :rtype: set"""

        # noinspection PyUnresolvedReferences
        def _last(s1):
            queue, done = {s1}, set()
            while queue:
                q = queue.pop()
                done.add(q)
                for k in self.delta.get(q, {}):
                    for s1 in forceIterable(self.delta[q][k]):
                        if self.finalP(s1):
                            return False
                        elif s1 not in done:
                            queue.add(s1)
            return True

        leaves = set()
        for s in self.Final:
            if _last(s):
                leaves.add(self.States[s])
        return leaves


class ADFA(fa.DFA, AFA):
    """Acyclic Deterministic Finite Automata class

    .. inheritance-diagram:: ADFA

    .. versionchanged:: 1.3.3
    """
    def __init__(self):
        fa.DFA.__init__(self)
        AFA.__init__(self)
        self.Minimal = False

    def __repr__(self):
        return 'ADFA({0:s})'.format(self.__str__())

    def complete(self, dead=None):
        """Make the ADFA complete

        :param int dead: a state to be identified as dead state if one was not identified yet
        :rtype: ADFA

        .. attention::
           The object is modified in place

        .. versionchanged:: 1.3.3"""
        if dead is not None:
            self.Dead = dead
        else:
            try:
                if self.Dead is None:
                    raise AttributeError
                else:
                    _ = self.Dead
            except AttributeError:
                foo = self.addState(DeadName)
                self.Dead = foo
        for st in range(len(self.States)):
            for k in self.Sigma:
                if k not in self.delta.get(st, {}).keys():
                    self.addTransition(st, k, self.Dead)
        self.Minimal = False
        return self

    def dup(self):
        """Duplicate the basic structure into a new ADFA. Basically a copy.deep.

        :rtype: ADFA"""
        return deepcopy(self)

    def __invert__(self):
        """ Complement of a ADFA is a DFA

        :return:DFA
        """
        aut = self.forceToDFA()
        return ~aut

    def minimalP(self, method=None):
        """Tests if the DFA is minimal

        :param method: minimization algorithm (here void)
        :rtype: bool

        .. versionchanged:: 1.3.3"""
        if self.Minimal:
            return True
        foo = self.minimal()
        if self.completeP():
            foo.complete()
        answ = len(foo) == len(self)
        if answ:
            self.Minimal = True
        return answ

    def forceToDFA(self):
        """ Conversion to DFA

        :rtype: DFA"""
        new = fa.DFA()
        new.States = deepcopy(self.States)
        new.Sigma = deepcopy(self.Sigma)
        new.Initial = self.Initial
        new.Final = copy(self.Final)
        for s in self.delta:
            for c in self.delta[s]:
                new.addTransition(s, c, self.delta[s][c])
        return new

    def forceToDFCA(self):
        """ Conversion to DFCA

        :rtype: DFA"""
        new = fa.DFA()
        new.States = deepcopy(self.States)
        new.Sigma = deepcopy(self.Sigma)
        new.Initial = self.Initial
        new.Final = copy(self.Final)
        for s in self.delta:
            for c in self.delta[s]:
                new.addTransition(s, c, self.delta[s][c])
        return new

    def wordGenerator(self):
        """Creates a random word generator

        :return: the random word generator
        :rtype: RndWGen

        .. versionadded:: 1.2"""
        return RndWGen(self)

    def possibleToReverse(self):
        """Tests if language is reversible

        .. versionadded:: 1.3.3"""
        return True

    def minimal(self):
        """Finds the minimal equivalent ADFA

        .. seealso:: [TCS 92 pp 181-189] Minimisation of acyclic deterministic automata in linear time, Dominique Revuz

        .. versionchanged:: 1.3.3

        :returns: the minimal equivalent ADFA
        :rtype: ADFA"""

        def _getListDelta(ss):
            """returns [([sons,final?],s) for s in ss].sort"""
            l = []
            for s in ss:
                dl = [new.delta[s][k] for k in new.Sigma]
                dl.append(s in new.Final)
                l.append((dl, s))
            l.sort()
            return l

        def _collapse(r1, r2):
            """redirects all transitions going to r2 to r1 and adds r2 to toBeDeleted"""
            for s in rdelta[r2]:
                for k in new.delta[s]:
                    if new.delta[s][k] == r2:
                        new.delta[s][k] = r1
            toBeDeleted.append(r2)

        if len(self.States) == 1:
            return self
        new = deepcopy(self)
        new.trim()
        new.complete()
        if new.Dead is None:
            deadName = None
        else:
            deadName = new.States[new.Dead]
        rank, rdelta = new.evalRank()
        toBeDeleted = []
        maxr = len(rank) - 2
        for r in xrange(maxr + 1):
            ls = _getListDelta(rank[r])
            (d0, s0) = ls[0]
            j = 1
            while j < len(ls):
                (d1, s1) = ls[j]
                if d0 == d1:
                    _collapse(s0, s1)
                else:
                    (d0, s0) = (d1, s1)
                j += 1
        new.deleteStates(toBeDeleted)
        if deadName is not None:
            new.Dead = new.stateIndex(deadName)
        new.Minimal = True
        return new

    def minReversible(self):
        """Returns the minimal reversible equivalent automaton

        :rtype: ADFA"""
        new = self.dup()
        rank, rdelta = new.evalRank()



    def statePairEquiv(self, s1, s2):
        """Tests if two states of a ADFA are equivalent

        :param int s1: state1
        :param int s2: state2
        :rtype: bool

        .. versionadded:: 1.3.3"""
        if not self.same_nullability(s1, s2):
            return False
        else:
            return self.delta.get(s1, {}) == self.delta.get(s2, {})

    def addSuffix(self, st, w):
        """Adds a suffix starting in st

        :param int st: state
        :param Word w: suffix

        .. versionadded:: 1.3.3

        .. attention:: in place transformation"""
        s1 = st
        for c in w:
            s2 = self.addState()
            self.addTransition(s1, c, s2)
            s1 = s2
        self.addFinal(s1)

    def _last_child(self, s):
        """to be used by xxx of FL.MADFA

        :param int s: state index
        :returns: pair state index / symbol

        .. versionadded:: 1.3.3"""
        for c in sorted(list(self.Sigma)).__reversed__():
            if c in self.delta.get(s, {}):
                return self.delta[s][c], c
        raise FAdoGeneralError("Something unexpected in _last_child({:d})".format(s))

    def _replace_or_register(self, s, r):
        """to be used by xxx of FL.MADFA

        :param int s: state
        :param Set r: register (inherited from context)

        .. versionadded:: 1.3.3"""
        (child, c) = self._last_child(s)
        if self.delta.get(child, {}):
            self._replace_or_register(child, r)
        for q in r:
            if self.statePairEquiv(q, child):
                self.delta[s][c] = q
                self.deleteState(child)
                return
        r.add(child)

    def _common_prefix(self, wrd):
        """The longest prefix of w that can be read in the ADFA and the correspondent state

        :param Word wrd: word"""
        pref = Word()
        q = self.Initial
        for s in wrd:
            if s in self.delta.get(q, {}):
                pref.append(s)
                q = self.delta[q][s]
            else:
                break
        return pref, q

    def _addWordToMinimal(self, w):
        """Incremental minimization algorithm

        :param Word w: word

        .. attention:: in place transformation

        .. versionadded:: 1.3.3

        .. seealso:: Incremental Construction of Minimal Acyclic Finite-State Automata, J.Daciuk, S.Mihov,
                     B.Watson and R.E.Watson"""

        def _transverseNonConfluence(wrd):
            inCount = dict()
            for s in range(len(self.States)):
                for c in self.delta.get(s, {}):
                    for s1 in self.delta[s][c]:
                        inCount[s1] = inCount.get(s1, 0) + 1
            q = self.Initial
            visited = [q]
            for i, sym in enumerate(wrd):
                if sym not in self.delta.get(s, {}) or inCount[self.delta[q][sym]] > 1:
                    return q, i
                q = self.delta[q][sym]
                visited.append(q)

        def _cloneConfluence(st, wrd, ind):
            q = st
            for i, sym in enumerate(wrd[ind:]):
                if sym not in self.delta.get(q, {}):
                    return q, i + ind
                qn = self.delta[q][sym]
                sn = self.addState()
                for c1 in self.delta.get(qn, {}):
                    self.delta.setdefault(sn, {})[c1] = self.delta[qn][c1]
                    if self.finalP(qn):
                        self.addFinal(sn)
                self.addTransition(q, sym, sn)
                q = sn
                visited.append(q)

        def _replOrReg(st, wrd):
            if len(w) != 0:
                self.addTransition(q,wrd[0],_replOrReg(self.delta[st][wrd[0]], wrd[1:]))
            else:
                for c in register:
                    if self.statePairEquiv(c, q):
                        self.deleteState(q)
                        return c
                register.add(q)

        def _addSuffix(st, wrd):
            s = st
            for c in wrd:
                sn = self.addState()
                self.addTransition(s, c, sn)
                s = sn
            self.addFinal(s)

        register = set()
        visited = []
        q, i = _transverseNonConfluence(w)
        f = q
        register.remove(q)
        j = i
        q, i = _cloneConfluence(q, w, i)
        _addSuffix(q, w[i:])
        if j < len(w):
            self.delta[f][w[j]] = _replOrReg(self.delta[f][w[j]], w[j+1:])



    def dissMin(self, witnesses=None):
        """Evaluates the minimal dissimilarity language
        :param dict witnesses: optional witness dictionay
        :rtype: FL

        .. versionadded:: 1.2.1"""
        new = self.minimal()
        sz = len(new.States)
        todo = [(i, j) for i in range(sz) for j in range(i)]
        mD = FL(Sigma=new.Sigma)
        lvl = new.level()
        rnk = new.directRank()
        l = max([rnk[x] for x in rnk])
        Li = []
        for (i, j) in todo:
            if self.finalP(i) ^ self.finalP(j):
                if witnesses is not None:
                    witnesses[(i, j)] = Word(Epsilon)
                Li.append((i, j))
                mD.addWord(Word(Epsilon))
        delFromList(todo, Li)
        words = self.words()
        for w in words:
            if len(w) >= l or not todo:
                break
            Li = []
            for (i, j) in todo:
                if (lvl[i] + len(w) > l) or (lvl[j] + len(w) > l):
                    Li.append((i, j))
                elif self.evalWordP(w, i) ^ self.evalWordP(w, j):
                    mD.addWord(w)
                    if witnesses is not None:
                        witnesses[(i, j)] = w
                    Li.append((i, j))
            delFromList(todo, Li)
        return mD

    def diss(self):
        """ Evaluates the dissimilarity language

        :rtype: FL

        .. versionadded:: 1.2.1"""
        new = self.minimal()
        n = len(new.States)
        mD = FL(Sigma=new.Sigma)
        lvl = new.level()
        rnk = new.directRank()
        l = max([rnk[x] for x in rnk])
        if len(new.Final) != n:
            mD.addWord(Word(Epsilon))
        words = self.words()
        for w in words:
            lw = len(w)
            if lw >= l:
                break
            skip = False
            for i in range(n):
                if skip:
                    break
                for j in range(i):
                    if (lvl[i] + lw <= l) and (lvl[j] + lw <= l) and (self.evalWordP(w, i) ^ self.evalWordP(w, j)):
                        mD.addWord(w)
                        skip = True
                        break
        return mD

    def level(self):
        """Computes the level  for each state

        :returns: levels of states
        :rtype: dict

        .. versionadded:: 0.9.8"""
        lvl = {}
        done, alvl = set(), [self.Initial]
        l = 0
        while alvl:
            nlvl = set()
            for i in alvl:
                lvl[i] = l
                done.add(i)
                for c in self.delta[i]:
                    j = self.delta[i][c]
                    if j not in done and j not in alvl:
                        nlvl.add(j)
            l += 1
            alvl = copy(nlvl)
        return lvl

    def _gap(self, l, lvl):
        """Computes the gap value for each pair of states.

        The automata is supposed to have its states named numerically in such way that the initial is zero

        :param int l: length of the longest word
        :param dict lvl: level of each state
        :returns: gap function
        :rtype: dict"""
        def _range(r, s):
            return l - max(lvl[r], lvl[s])
        gp = {}
        n = len(self.States) - 1
        for i in range(n):
            gp[(self.stateIndex(i), self.stateIndex(n))] = l
        if lvl[self.stateIndex(n)] <= l:
            for i in self.Final:
                gp[(i, self.stateIndex(n))] = 0
        for i in range(n):
            for j in range(i + 1, n):
                if not self.same_nullability(self.stateIndex(i), self.stateIndex(j)):
                    gp[(self.stateIndex(i), self.stateIndex(j))] = 0
                else:
                    gp[(self.stateIndex(i), self.stateIndex(j))] = l
        for i in range(n - 2, -1, -1):
            for j in range(n, i, -1):
                for c in self.Sigma:
                    i1, j1 = self.delta[self.stateIndex(i)][c], self.delta[self.stateIndex(j)][c]
                    if i1 != j1:
                        if int(self.States[i1]) < int(self.States[j1]):
                            g = gp[(i1, j1)]
                        else:
                            g = gp[(j1, i1)]
                        if g + 1 <= _range(self.stateIndex(i), self.stateIndex(j)):
                            gp[(self.stateIndex(i), self.stateIndex(j))] = min(gp[(self.stateIndex(i),
                                                                                   self.stateIndex(j))], g + 1)
        return gp

    def minDFCA(self):
        """Generates a minimal deterministic cover automata from a DFA

        :rtype: DFCA

        .. versionadded:: 0.9.8

        .. seealso::
            Cezar Campeanu, Andrei Päun, and Sheng Yu, An efficient algorithm for constructing minimal cover
            automata for finite languages, IJFCS"""
        new = self.dup().minimal()
        if not self.completeP():
            new.complete()
        rank = new.directRank()
        irank = dict((v, [k for (k, xx) in filter(lambda (key, value): value == v, rank.items())])
                     for v in set(rank.values()))
        l = rank[new.Initial]
        lvl = new.level()
        foo = [x for x in irank]
        foo.sort(reverse=True)
        lnames = [None for _ in new.States]
        newname = 0
        for i in foo:
            for j in irank[i]:
                lnames[j] = newname
                newname += 1
        new.States = lnames
        g = new._gap(l, lvl)
        P = [False for _ in new.States]
        toMerge = []
        for i in range(len(new.States) - 1):
            if not P[i]:
                for j in range(i + 1, len(new.States)):
                    if not P[j] and g[(new.stateIndex(i), new.stateIndex(j))] == l:
                        toMerge.append((i, j))
                        P[j] = True
        for (a, b) in toMerge:
            new.mergeStates(new.stateIndex(b), new.stateIndex(a))
        new.trim()
        new = new.forceToDFCA()
        new.length = l
        return new

    def trim(self):
        """Remove states that do not lead to a final state, or, inclusively, that can't be reached from the initial
        state. Only useful states remain.

        .. attention:: in place transformation"""
        fa.OFA.trim(self)
        try:
            del self.Dead
        except AttributeError:
            pass
        return self

    def toANFA(self):
        """Converts the ADFA in a equivalent ANFA

        :rtype: ANFA"""
        new = ANFA()
        new.setSigma(copy(self.Sigma))
        new.States = copy(self.States)
        for s in xrange(len(self.States)):
            for k in self.delta.get(s, {}):
                new.addTransition(s, k, self.delta[s][k])
        new.addInitial(self.Initial)
        for s in self.Final:
            new.addFinal(s)
        return new

    def toNFA(self):
        """Converts the ADFA in a equivalent NFA

        :rtype: ANFA

        .. versionadded:: 1.2"""
        return self.toANFA()


class RndWGen(object):
    """Word random generator class

    .. versionadded:: 1.2"""
    def __init__(self, aut):
        """
        :param aut: automata recognizing the language
        :type aut: ADFA """
        self.Sigma = list(aut.Sigma)
        self.table = dict()
        self.aut = aut.minimal()
        rank, _ = self.aut.evalRank()
        self.aut._compute_delta_inv()
        deltai = self.aut.delta_inv
        mrank = max(rank)
        for i in range(0, mrank + 1):
            for s in rank[i]:
                self.table.setdefault(s, {})
                if self.aut.finalP(s):
                    final = 1
                else:
                    final = 0
                self.table[s][None] = sum([self.table[s].get(c, 0) for c in self.Sigma])
                for c in self.Sigma:
                    rs = deltai[s].get(c, [])
                    for r in rs:
                        self.table.setdefault(r, {})
                        self.table[r][c] = self.table[s][None] + final

    @staticmethod
    def _rndChoose(l):
        sm = sum(l)
        r = random.randint(1, sm)
        for i, j in enumerate(l):
            if r <= j:
                return i
            else:
                r -= j

    def next(self):
        """Next word

        :return: a new random word"""
        word = Word()
        s = self.aut.Initial
        while True:
            if self.aut.finalP(s) and random.randint(1, self.table[s][None] + 1) == 1:
                return word
            i = self._rndChoose([self.table[s].get(c, 0) for c in self.Sigma])
            word.append(self.Sigma[i])
            s = self.aut.delta[s][self.Sigma[i]]


# noinspection PyUnresolvedReferences
class ANFA(fa.NFA, AFA):
    """Acyclic Nondeterministic Finite Automata class

    .. inheritance-diagram:: ANFA"""

    def moveFinal(self, st, stf):
        """Unsets a set as final transfering transition to another final
        :param int st: the state to be 'moved'
        :param int stf: the destination final state

        .. note::
           stf must be a 'last' final state, i.e., must have no out transitions to anywhere but to a possible dead
           state

        .. attention:: the object is modified in place"""
        (rdelta, _) = self._getRdelta()
        for s in rdelta[st]:
            l = []
            for k in self.delta[s]:
                if st in self.delta[s][k]:
                    l.append(k)
            for k in l:
                self.addTransition(s, k, stf)
            self.delFinal(s)

    def mergeStates(self, s1, s2):
        """Merge state s2 into state s1

        :param int s1: state
        :param int s2: state

        .. note::
           no attempt is made to check if the merging preserves the language of teh automaton

        .. attention:: the object is modified in place"""
        (_, rdelta) = self._getRdelta()
        for s in rdelta[s2]:
            l = []
            for k in self.delta[s]:
                if s2 in self.delta[s][k]:
                    l.append(k)
            for k in l:
                self.delta[s][k].remove(s2)
                self.addTransition(s, k, s1)
        for k in self.delta.get(s2, {}):
            for ss in self.delta[s2][k]:
                self.delta.setdefault(s1, {}).setdefault(k, set()).add(ss)
        self.deleteState(s2)

    def mergeLeaves(self):
        """Merge leaves

        .. attention:: object is modified in place"""
        l = self.getLeaves()
        if len(l):
            s0n = l.pop()
            while l:
                s0 = self.stateIndex(s0n)
                s = self.stateIndex(l.pop())
                self.mergeStates(s0, s)

    def mergeInitial(self):
        """Merge initial states

        .. attention:: object is modified in place"""
        l = copy(self.Initial)
        s0 = self.stateIndex(l.pop())
        while l:
            s = self.stateIndex(l.pop())
            self.mergeStates(s0, s)


def sigmaInitialSegment(Sigma, l, exact=False):
    """Generates the ADFA recognizing Sigma^i for i<=l
    :param set Sigma: the alphabet
    :param int l: length
    :param bool exact: only the words with exactly that length?
    :returns: the automaton
    :rtype: ADFA"""
    new = ADFA()
    new.setSigma(Sigma)
    s = new.addState()
    if not exact:
        new.addFinal(s)
    new.setInitial(s)
    for i in range(l):
        s1 = new.addState()
        if not exact or i == l - 1:
            new.addFinal(s1)
        for k in Sigma:
            new.addTransition(s, k, s1)
        s = s1
    return new


# noinspection PyUnboundLocalVariable
def genRndTrieBalanced(maxL, Sigma, safe=True):
    """Generates a random trie automaton for a binary language of balanced words of a given leght for max word
    :param int maxL: length of the max word
    :param set Sigma: alphabet to be used
    :param bool safe: should a word of size maxl be present in every language?
    :return: the generated trie automaton
    :rtype: ADFA"""

    def _genEnsurance(m, alphabet):
        l = len(alphabet)
        fair = m / l
        if m % l == 0:
            odd = 0
        else:
            odd = 1
        pool = copy(alphabet)
        c = {}
        sl = []
        while len(sl) < m:
            s1 = random.choice(pool)
            c[s1] = c.get(s1, 0) + 1
            if c[s1] == fair + odd:
                pool.remove(s1)
            sl.append(s1)
        return sl

    def _legal(cont):
        l = [cont[k1] for k1 in cont]
        return max(l) - min(l) <= 1

    def _descend(s1, ens, safe1, m, cont):
        sons = 0
        if not safe1:
            if _legal(cont):
                final = random.randint(0, 1)
            else:
                final = 0
        # noinspection PyUnboundLocalVariable
        if safe1:
            trie.addFinal(s1)
            final = 1
        elif final == 1:
            trie.addFinal(s1)
        if m != 0:
            if safe1:
                ks = ens.pop()
            else:
                ks = None
            for k1 in trie.Sigma:
                ss = trie.addState()
                trie.addTransition(s1, k1, ss)
                cont[k1] = cont.get(k1, 0) + 1
                if _descend(ss, ens, k1 == ks, m - 1, cont):
                    sons += 1
                cont[k1] -= 1
        if sons == 0 and final == 0:
            trie.deleteState(s1)
            return False
        else:
            return True

    if safe:
        ensurance = _genEnsurance(maxL, Sigma)
    else:
        ensurance = None
    trie = ADFA()
    trie.setSigma(Sigma)
    s = trie.addState()
    trie.setInitial(s)
    contab = {}
    for k in Sigma:
        contab[k] = 0
    _descend(s, ensurance, safe, maxL, contab)
    if random.randint(0, 1) == 1:
        trie.delFinal(s)
    return trie


# noinspection PyUnboundLocalVariable
def genRndTrieUnbalanced(maxL, Sigma, ratio, safe=True):
    """Generates a random trie automaton for a binary language of balanced words of a given length for max word

    :param int maxL: length of the max word
    :param set Sigma: alphabet to be used
    :param int ratio: the ratio of the unbalance
    :param bool safe: should a word of size maxl be present in every language?
    :return: the generated trie automaton
    :rtype: ADFA"""

    def _genEnsurance(m, alphabet):
        chief = uSet(alphabet)
        fair = m / (ratio + 1)
        pool = list(copy(alphabet))
        c = {}
        sl = []
        while len(sl) < m:
            s1 = random.choice(pool)
            c[s1] = c.get(s1, 0) + 1
            if len(sl) - c.get(chief, 0) == fair:
                pool = [chief]
            sl.append(s1)
        return sl

    def _legal(cont):
        l = [cont[k1] for k1 in cont]
        return (ratio + 1) * cont[uSet(Sigma)] >= sum(l)

    # noinspection PyUnboundLocalVariable
    def _descend(s1, ens, safe1, m, cont):
        sons = 0
        if not safe1:
            if _legal(cont):
                final = random.randint(0, 1)
            else:
                final = 0
        if safe1:
            trie.addFinal(s1)
            final = 1
        elif final == 1:
            trie.addFinal(s1)
        if m:
            if safe1:
                ks = ens.pop()
            else:
                ks = None
            for k1 in trie.Sigma:
                ss = trie.addState()
                trie.addTransition(s1, k1, ss)
                cont[k1] = cont.get(k1, 0) + 1
                if _descend(ss, ens, k1 == ks, m - 1, cont):
                    sons += 1
                cont[k1] -= 1
        if sons == 0 and final == 0:
            trie.deleteState(s1)
            return False
        else:
            return True

    if safe:
        ensurance = _genEnsurance(maxL, Sigma)
    else:
        ensurance = None
    trie = ADFA()
    trie.setSigma(Sigma)
    s = trie.addState()
    trie.setInitial(s)
    contab = {}
    for k in Sigma:
        contab[k] = 0
    _descend(s, ensurance, safe, maxL, contab)
    if random.randint(0, 1) == 1:
        trie.delFinal(s)
    return trie


# noinspection PyUnboundLocalVariable
def genRandomTrie(maxL, Sigma, safe=True):
    """Generates a random trie automaton for a finite language with a given length for max word
    :param int maxL: length of the max word
    :param set Sigma: alphabet to be used
    :param bool safe: should a word of size maxl be present in every language?
    :return: the generated trie automaton
    :rtype: ADFA"""

    def _genEnsurance(m, alphabet):
        l = len(alphabet)
        sl = list(alphabet)
        return [sl[random.randint(0, l - 1)] for _ in xrange(m)]

    # noinspection PyUnboundLocalVariable
    def _descend(s1, ens, safe1, m):
        sons = 0
        final = None
        if not safe1:
            final = random.randint(0, 1)
        if safe1:
            trie.addFinal(s1)
            final = 1
        elif final == 1:
            trie.addFinal(s1)
        if m:
            if safe1:
                ks = ens.pop()
            else:
                ks = None
            for k in trie.Sigma:
                ss = trie.addState()
                trie.addTransition(s1, k, ss)
                if _descend(ss, ens, k == ks, m - 1):
                    sons += 1
        if sons == 0 and final == 0:
            trie.deleteState(s1)
            return False
        else:
            return True

    if safe:
        ensurance = _genEnsurance(maxL, Sigma)
    else:
        ensurance = None
    trie = ADFA()
    trie.setSigma(Sigma)
    s = trie.addState()
    trie.setInitial(s)
    _descend(s, ensurance, safe, maxL)
    if random.randint(0, 1) == 1:
        trie.delFinal(s)
    return trie


# noinspection PyUnboundLocalVariable
def genRndTriePrefix(maxL, Sigma, ClosedP=False, safe=True):
    """Generates a random trie automaton for a finite (either prefix free or prefix closed) language with a given
    length for max word
    :param int maxL: length of the max word
    :param set Sigma: alphabet to be used
    :param bool ClosedP: should it be a prefix closed language?
    :param bool safe: should a word of size maxl be present in every language?
    :return: the generated trie automaton
    :rtype: ADFA"""

    def _genEnsurance(m, alphabet):
        l = len(alphabet)
        sl = list(alphabet)
        return [sl[random.randint(0, l - 1)] for _ in xrange(m)]

    def _descend(s1, ens, saf, m):
        sons = ClosedP
        if m is 0:
            final = random.randint(0, 1)
            if saf or final == 1:
                trie.addFinal(s1)
                return True
            else:
                return False
        else:
            if saf is True:
                ks = ens.pop()
            else:
                ks = None
            for k in trie.Sigma:
                ss = trie.addState()
                trie.addTransition(s1, k, ss)
                r = _descend(ss, ens, k == ks, m - 1)
                if not ClosedP:
                    sons |= r
                else:
                    sons &= 1
            if not ClosedP:
                if not sons:
                    final = random.randint(0, 1)
                    if final == 1:
                        trie.addFinal(s1)
                        return True
                    else:
                        return False
                else:
                    return True
            else:
                if not sons:
                    final = random.randint(0, 1)
                    if final == 1:
                        trie.addFinal(s1)
                        return True
                    else:
                        return False
                else:
                    trie.addFinal(s1)
                    return True

    ensurance = None
    if safe:
        ensurance = _genEnsurance(maxL, Sigma)
    trie = ADFA()
    trie.setSigma(Sigma)
    s = trie.addState()
    trie.setInitial(s)
    _descend(s, ensurance, safe, maxL)
    return trie


def DFAtoADFA(aut):
    """Transforms an acyclic DFA into a ADFA

    :param DFA aut: the automaton to be transformed
    :raises notAcyclic: if the DFA is not acyclic
    :returns: the converted automaton
    :rtype: ADFA"""
    new = deepcopy(aut)
    new.trim()
    if not new.acyclicP(True):
        raise notAcyclic()
    afa = ADFA()
    afa.States = copy(new.States)
    afa.Sigma = copy(new.Sigma)
    afa.Initial = new.Initial
    afa.delta = copy(new.delta)
    afa.Final = copy(new.Final)
    afa.complete()
    return afa


def stringToADFA(s):
    """Convert a canonical string representation of a ADFA to a ADFA
    :param list s: the string in its canonical order
    :returns: the ADFA
    :rtype: ADFA

    .. seealso::
        Marco Almeida, Nelma Moreira, and Rogério Reis. Exact generation of minimal acyclic deterministic finite
        automata. International Journal of Foundations of Computer Science, 19(4):751-765, August 2008.
    """
    k = len(s[0]) - 1
    new = ADFA()
    new.setSigma([str(c) for c in range(k)])
    for st, sts in enumerate(s):
        new.addState(str(st))
        for c, s1 in enumerate(sts[:-1]):
            new.addTransition(st, str(c), s1)
        if sts[-1]:
            new.addFinal(st)
    new.setInitial(len(s) - 1)
    return new




# coding=utf-8
"""**Random DFA generation (alternative version in python)**

ICDFA Random generation binding

.. *Authors:* Rogério Reis & Nelma Moreira

.. *This is part of FAdo project*  http://fado.dcc.fc.up.pt

.. *Version:* 1.0

.. *Copyright:* 1999-2014 Rogério Reis & Nelma Moreira {rvr,nam}@dcc.fc.up.pt

.. versionadded:: 1.0

..  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA."""

import random
import time
import common
import fa


class ICDFArgen(object):
    """Generic ICDFA random generator class

    :var int n: number of states
    :var int k: size of the alphabet
    :var int pn: how more problable shall a non defined transition be?
    :var int seed: seed for the random generator. Default is to generate a time & system dependent.

    .. seealso:: Marco Almeida, Nelma Moreira, and Rogério Reis. Enumeration and generation with a string automata
       representation. Theoretical Computer Science, 387(2):93-102, 2007

    .. versionchanged:: 1.3.4 seed added to the random generator"""

    def __init__(self, n, k, nd=False, pn=1, seed=0):
        self.N = dict()
        self.n = n
        self.k = k
        if seed != 0:
            self.seed = seed
        else:
            self.seed = hash(time.clock())
        foo = n ** k
        random.seed(self.seed)
        if not nd:
            self.pn = 0
        else:
            self.pn = pn
        for j in range((n - 1) * k - 1, n - 3, -1):
            self.N[(n - 1, j)] = foo
            foo *= n
        for m in range(n - 2, 0, -1):
            foo = 0
            bar = 1
            m1 = m + 1
            for i in range(0, k):
                foo += bar * self.N[(m + 1, m * k + i)]
                bar *= m1
            self.N[(m, m * k - 1)] = foo
            for i in range(m * k - 2, m - 2, -1):
                self.N[(m, i)] = m1 * self.N[(m, i + 1)] + self.N[(m + 1, i + 1)]

    def __iter__(self):
        return self

    def genFinalities(self):
        """ Generate bit map of final states

        :rtype: list
        """
        return [random.randint(0, 1) for _ in range(self.n)]

    def _getFlag(self, m, l):
        k = self.k
        bar = 1
        foo = 0
        for i in range(l, m * k):
            foo += bar * self.N[(m, i)]
            bar *= m
        r = random.randint(0, foo)
        bar = 1
        for i in range(l, m * k):
            foo = bar * self.N[(m, i)]
            if r < foo:
                return i
            else:
                r -= foo
                bar *= m
        return m * k - 1

    def _rndT(self, i):
        r = random.randint(0, i + self.pn - 1)
        if r < self.pn:
            return -1
        else:
            return r - self.pn

    def next(self):
        """ Generate an ICDFA
        """
        g = -1
        s = []
        for i in range(1, self.n):
            flag = self._getFlag(i, g + 1)
            for j in range(g + 1, flag):
                s.append(self._rndT(i))
            s.append(i)
            g = flag
        for i in range(g + 1, self.n * self.k):
            s.append(self._rndT(self.n))
        return fa.stringToDFA(s, self.genFinalities(), self.n, self.k)


class ICDFArnd(ICDFArgen):
    """ Complete ICDFA random generator class

    This is the class for the uniform random generator for Initially Connected DFAs

    :var int n: number of states
    :var int k: size of alphabet
    :var int seed: seed for the random generator (if 0 uses time as seed)

    .. note::
        This is an abstract class, not to be used directly

    .. versionchanged:: 1.3.4 seed added to the random generator"""

    def __init__(self, n, k, seed=0):
        super(ICDFArnd, self).__init__(n, k, False, 1, seed)

    def __str__(self):
        return "ICDFArnd %d %d %d" % (self.n, self.k, self.seed)


class ICDFArndIncomplete(ICDFArgen):
    """ Incomplete ICDFA random generator class

    :var int n: number of states
    :var int k: size of alphabet
    :var float bias: how often must the gost sink state appear (default None)
    :var int seed: seed for the random generator (if 0 uses time as seed)


    :raises IllegalBias: if a bias >=1 or <=0 is provided

    .. versionchanged:: 1.3.4 seed added to the random generator"""

    def __init__(self, n, k, bias=None, seed=0):
        if bias is None:
            super(ICDFArndIncomplete, self).__init__(n, k, True, 1, seed)
        elif bias <= 0 or bias >= 1:
            raise common.IllegalBias()
        else:
            m = int((bias * n) / (1 - bias))
            super(ICDFArndIncomplete, self).__init__(n, k, True, m, seed)
        self.n, self.k, self.bias = n, k, bias

    def __str__(self):
        return "ICDFArndIncomplete %d %d %d %d" % (self.n, self.k, self.bias, self.seed)


# -*- coding: utf-8 -*-
""" Witness language families for operations on DFAs

@author: Rogério Reis & Nelma Moreira

This is part of U{FAdo project <http://www.ncc.up.pt/FAdo>}.

Deterministic and non-deterministic automata manipulation, conversion and
evaluation.

@copyright: 1999-2012 Rogério Reis & Nelma Moreira {rvr,nam}@ncc.up.pt

Contributions by
 - Marco Almeida
 - Hugo Gouveia 
 - Davide Nabais
 - Eva Maia

B{Naming convention:} methods suffixed by P have boolean return.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA."""

from comboperations import *
from reex import *
from fa import DFA

# Useful automata


def emptyDFA(sigma=None):
    """
    Returns the minimal DFA for emptyset (incomplete)

    :param sigma:
    :return:
    """
    d = DFA()
    if sigma is not None:
        d.setSigma(sigma)
    i = d.addState()
    d.setInitial(i)
    return d


def epsilonDFA(sigma=None):
    """
    Returns the minimal DFA for {epsilon} (incomplete)

    :param sigma:
    :return:
    """
    d = DFA()
    if sigma is not None:
        d.setSigma(sigma)
    i = d.addState()
    d.setInitial(i)
    d.addFinal(i)
    return d

### Worst case automata for each operation

witnessDFA = {"toDFA": [("toDFAWCMF", "int"),
                        ("toDFAWC2", "int"),
                        ("toDFAWC3", "int")],
              "reversal": [("reversalWC3M", "int"),
                           ("reversalMB", "int"),
                           ("reversalWC3L", "int"),
                           ("reversalternaryWC", "int"),
                           ("reversalbinaryWC", "int")],
              "star": [("starWC", "int"),
                       ("starWCM", "int")],
              "concat": [("concatWC", "int", "int"),
                         ("concatWCM", "int", "int")],
              "conjunction": [("interWC", "int", "int")],
              "__or__": [("disjWC", "int", "int")],
              "shuffle": [("shuffleWC", "int", "int")],
              "suff": [("suffWCe","int"),("suffWCd","int")],
              "starDisj": [("starDisjWC", "int", "int")],
              "starInter": [("starInterBC", "int", "int")],
              "disjWStar": [("disjWStarWC", "int", "int")]}


def toDFAWC2MF(m=5):
    """ Worst case automata for toDFA(NFA) with n > 2, k=2
    ..seealso::
    A. R. Meyer and M. J. Fischer. Economy of description by automata,
    grammars, and formal systems. Twelfth Annual Symposium on
    Switching and Automata Theory, 1971,  188–191. IEEE Society Press.

    :param m: number of states
    :type m: integer
    :return: a dfa
    :rtype: DFA"""

    if m < 3:
        raise TestsError("number of states must be greater than 2")
    f = NFA()
    f.setSigma(["a", "b"])
    f.States = range(m)
    f.setInitial([0])
    f.addFinal(0)
    f.addTransition(0, "a", 1)
    for i in range(1, m):
        f.addTransition(i, "a", (i + 1) % m)
        f.addTransition(i, "b", i)
        f.addTransition(i, "b", 0)
    return f


def toDFAWC2(m=5):
    """ Worst case automata for toDFA(NFA) with n > 2, k=2
    ..seealso:: F.R. Moore. On the bounds for state-set size in the proofs
    of equivalence between deterministic, nondeterministic, and
    two-way finite automata. IEEE Transactions on computers, 2:1211–1214, 1971.

    :arg m: number of states
    :type m: integer
    :returns: a dfa
    :rtype: DFA"""

    if m < 3:
        raise TestsError("number of states must be greater than 2")
    f = NFA()
    f.setSigma(["a", "b"])
    f.States = range(m)
    f.setInitial([0])
    f.addFinal(m - 1)
    f.addTransition(0, "a", 1)
    f.addTransition(0, "b", 0)
    f.addTransition(m - 1, "a", 0)
    f.addTransition(m - 1, "a", 1)
    for i in range(1, m - 1):
        f.addTransition(i, "a", i + 1)
        f.addTransition(i, "b", i + 1)
    return f


def toDFAWC3(m=5):
    """ Worst case automata for toDFA(NFA) with n > 2, k=3.
    ..seealso:: O. B. Lupanov. A comparison of two types of finite sources.
    Problemy Kibernetiki, 9:321–326, 1963.
    
    :arg m: number of states
    :type m: integer
    :returns: a dfa
    :rtype: DFA"""

    if m < 3:
        raise TestsError("number of states must be greater than 2")
    f = NFA()
    f.setSigma(["a", "b", "c"])
    f.States = range(m)
    f.setInitial([0])
    f.addFinal(0)
    f.addTransition(0, "a", 1)
    f.addTransition(0, "b", 1)
    f.addTransition(1, "b", 0)
    f.addTransition(1, "c", 0)
    f.addTransition(1, "c", 1)
    f.addTransition(1, "a", 2)
    f.addTransition(m - 1, "a", 0)
    f.addTransition(m - 1, "b", m - 1)
    f.addTransition(m - 1, "c", m - 1)
    for i in range(2, m - 1):
        f.addTransition(i, "a", i + 1)
        f.addTransition(i, "b", i)
        f.addTransition(i, "c", i)
    return f


def reversalWC3M(m=5):
    """ Worst case automata for reversal(DFA) with m > 2, k=3.

    ..seealso:: Boris G. Mirkin. On dual automata. Kibernetika, 2:7–10, 1966.
   
    :arg m: number of states
    :type m: integer
    :returns: a dfa
    :rtype: DFA"""

    if m < 3:
        raise TestsError("number of states  must be greater than 2")
    return toDFAWC3(m).reversal()


def starSC(m=5):
    """ Worst case state complexity for star
    :arg m: number of states
    :type m: integer
    :returns: state complexity
    :rtype: integer"""

    if m > 1:
        return 3 * 2 ** (m - 2)
    return 1


def starWC(m=5):
    """ Worst case automata for star(DFA) with m > 2, k=2
    ..seealso:: S. Yu, Q. Zhuang, and K. Salomaa. The state complexities
    of some basic operations on regular languages.
    Theor. Comput. Sci., 125(2):315–328, 1994.

    :arg m: number of states
    :type m: integer
    :returns: a dfa
    :rtype: DFA"""

    if m < 3:
        raise TestsError("number of states must be greater than 2")
        # for m=2, L=\{w\in\{a,b\}*| |w|a odd \}
    f = DFA()
    f.setSigma(["a", "b"])
    f.States = range(m)
    f.setInitial(0)
    f.addFinal(m - 1)
    f.addTransition(0, "a", 1)
    f.addTransition(0, "b", 0)
    for i in range(1, m):
        f.addTransition(i, "a", (i + 1) % m)
        f.addTransition(i, "b", (i + 1) % m)
    return f


def starWCM(m=5):
    """ Worst case automata for star(DFA) with m > 2, k=2

    ..seealso:: A. N. Maslov. Estimates of the number of states of
    finite automata. Dokllady Akademii Nauk SSSR, 194:1266–1268, 1970. 
    
    :arg m: number of states
    :type m: integer
    :returns: a dfa
    :rtype: DFA"""

    if m < 3:
        raise TestsError("number of states must be greater than 2")
    f = DFA()
    f.setSigma(["a", "b"])
    f.States = range(m)
    f.setInitial(0)
    f.addFinal(m - 1)
    f.addTransition(m - 1, "a", 0)
    f.addTransition(m - 1, "b", m - 2)
    f.addTransition(0, "b", 0)
    f.addTransition(0, "a", 1)
    for i in range(1, m - 1):
        f.addTransition(i, "a", (i + 1))
        f.addTransition(i, "b", (i - 1))
    return f


def concatSC(m, n, k=1):
    """Worst case state complecity for concatenation
    :arg m: number of states
    :arg n: number of states
    :arg k: number of letters
    :type m: integer
    :type n: integer
    :type k: integer
    :returns: state compelxity
    :rtype: integer"""

    return m * 2 ** n - k * 2 ** (n - 1)


def concatWCM(m=4, n=4):
    """ Worst case automata for catenation(DFA,DFA) with m,n > 1, k=2,

    ..seealso:: A. N. Maslov. Estimates of the number of states of
    finite automata. Dokllady Akademii Nauk SSSR, 194:1266–1268, 1970. 
    :arg m: number of states
    :arg n: number of states
    :type m: integer
    :type n: integer
    :returns: two dfas 
    :rtype: (DFA, DFA)"""

    if n < 2 or m < 2:
        raise TestsError("number of states must be both greater than 1")
    d1, d2 = DFA(), DFA()
    d1.setSigma(["a", "b"])
    d1.States = range(m)
    d1.setInitial(0)
    d1.addFinal(m - 1)
    d1.addTransition(m - 1, "b", 0)
    d1.addTransition(m - 1, "a", m - 1)
    for i in range(m - 1):
        d1.addTransition(i, "a", i)
        d1.addTransition(i, "b", i + 1)
    d2.setSigma(["a", "b"])
    d2.States = range(n)
    d2.setInitial(0)
    d2.addFinal(n - 1)
    d2.addTransition(n - 1, "a", n - 1)
    d2.addTransition(n - 1, "b", n - 2)
    d2.addTransition(n - 2, "b", n - 1)
    d2.addTransition(n - 2, "a", n - 1)
    for i in range(n - 2):
        d2.addTransition(i, "a", i + 1)
        d2.addTransition(i, "b", i)

    return d1, d2


def concatWC(m=6, n=6):
    """ Worst case automata for catenation(DFA,DFA) with m,n > 1
    ..seealso:: S. Yu, Q. Zhuang, and K. Salomaa. The state complexities
    of some basic operations on regular languages.
    Theor. Comput. Sci., 125(2):315–328, 1994.
    :arg m: number of states
    :arg n: number of states
    :type m: integer
    :type n: integer
    :returns: two dfas 
    :rtype: (DFA, DFA)"""
    if n < 2 or m < 2:
        raise TestsError("number of states must both  greater than 1")
    d1, d2 = DFA(), DFA()
    d1.setSigma(["a", "b", "c"])
    d1.States = range(m)
    d1.setInitial(0)
    d1.addFinal(m - 1)
    for i in range(m):
        d1.addTransition(i, "a", (i + 1) % m)
        d1.addTransition(i, "b", 0)
        d1.addTransition(i, "c", i)
    d2.setSigma(["a", "b", "c"])
    d2.States = range(n)
    d2.setInitial(0)
    d2.addFinal(n - 1)
    for i in range(n):
        d2.addTransition(i, "b", (i + 1) % n)
        d2.addTransition(i, "a", i)
        d2.addTransition(i, "c", 1)
    return d1, d2

def concatWCB(m=4, n=4):
    """ Worst case automata for catenation(DFA,DFA) with m,n > 1, k=2,

    ..seealso::Jirásek, J., Jiráaskováa, G., Szabari, A., 2005.
     State complexity of concatenation and complementation of regular
     languages. Int. J. Found. Comput. Sci. 16 (3), 511–529.
    :arg m: number of states
    :arg n: number of states
    :type m: integer
    :type n: integer
    :returns: two dfas
    :rtype: (DFA, DFA)"""

    if n < 2 or m < 2:
        raise TestsError("number of states must be both greater than 1")
    d1, d2 = DFA(), DFA()
    d1.setSigma(["a", "b"])
    d1.States = range(m)
    d1.setInitial(0)
    d1.addFinal(m - 1)
    d1.addTransition(m - 1, "b", 0)
    d1.addTransition(m - 1, "a", m - 1)
    for i in range(m - 1):
        d1.addTransition(i, "a", i)
        d1.addTransition(i, "b", i + 1)
    d2.setSigma(["a", "b"])
    d2.States = range(n)
    d2.setInitial(0)
    d2.addFinal(n - 1)
    d2.addTransition(n - 1, "a", 0)
    d2.addTransition(n - 1, "b", 0)
    d2.addTransition(0, "a", 0)
    d2.addTransition(0, "b", 1)
    d2.addTransition(n - 2, "a", n - 1)
    for i in range(1,n - 1):
        d2.addTransition(i, "a", i + 1)
        d2.addTransition(i, "b", i + 1)

    return d1, d2


def interWC(m=6, n=5):
    """ Worst case automata for intersection(DFA,DFA) with m,n >1

    ..seealso:: S. Yu, Q. Zhuang, and K. Salomaa. The state complexities
    of some basic operations on regular languages.
    Theor. Comput. Sci., 125(2):315–328, 1994.
    :arg m: number of states
    :arg n: number of states
    :type m: integer
    :type n: integer
    :returns: two dfas 
    :rtype: (DFA, DFA)"""
    if n < 2 or m < 2:
        raise TestsError("number of states must be both greater than 1")
    d1, d2 = DFA(), DFA()
    d1.setSigma(["a", "b"])
    d1.States = range(m)
    d1.setInitial(0)
    d1.addFinal(0)
    for i in range(m):
        d1.addTransition(i, "a", (i + 1) % m)
        d1.addTransition(i, "b", i)
    d2.setSigma(["a", "b"])
    d2.States = range(m)
    d2.setInitial(0)
    d2.addFinal(0)
    for i in range(n):
        d2.addTransition(i, "b", (i + 1) % n)
        d2.addTransition(i, "a", i)
    return d1, d2


def disjWC(m=6, n=5):
    """ Worst case automata for disjunction(DFA,DFA) with m,n >1
    ..seealso:: S. Yu, Q. Zhuang, and K. Salomaa. The state complexities
    of some basic operations on regular languages.
    Theor. Comput. Sci., 125(2):315–328, 1994.
    :arg m: number of states
    :arg n: number of states
    :type m: integer
    :type n: integer
    :returns: two dfas 
    :rtype: (DFA, DFA)"""

    if n < 2 or m < 2:
        raise TestsError("number of states must be both greater than 1")
    d1, d2 = DFA(), DFA()
    d1.setSigma(["a", "b"])
    d1.States = range(m)
    d1.setInitial(0)
    d1.addTransition(0, "a", 1)
    d1.addTransition(0, "b", 0)
    for i in range(1, m):
        d1.addTransition(i, "a", (i + 1) % m)
        d1.addTransition(i, "b", i)
        d1.addFinal(i)
    d2.setSigma(["a", "b"])
    d2.States = range(m)
    d2.setInitial(0)
    d2.addTransition(0, "b", 1)
    d2.addTransition(0, "a", 0)
    for i in range(n):
        d2.addTransition(i, "b", (i + 1) % n)
        d2.addTransition(i, "a", i)
    d2.addFinal(0)
    return d1, d2


def reversalMB(m=8):
    """Worst case automata for reversal(DFA)

    ..seealso:: S. Yu, Q. Zhuang, and K. Salomaa. The state complexities
    of some basic operations on regular languages.
    Theor. Comput. Sci., 125(2):315–328, 1994.
    :arg m: number of states
    :type m: integer
    :returns: a dfa
    :rtype: DFA"""
    if m < 3:
        raise TestsError("number of states must be greater than 2")
    d = DFA()
    d.setSigma(["a", "b"])
    d.States = range(m)
    d.setInitial(0)
    for i in range(m):
        if i == m - 1:
            d.addTransition(m - 1, "a", 0)
        else:
            d.addTransition(i, "a", i + 1)
        if i == 2:
            d.addTransition(2, "b", 0)
        elif i == 3:
            d.addTransition(3, "b", 2)
        else:
            d.addTransition(i, "b", i)
    return d


def reversalWC3L(m=5):
    """ Worst case automata for reversal(DFA) with m > 2, k=3

    ..seealso:: E. L. Leiss. Succinct representation of regular languages
        by boolean automata ii. Theor. Comput. Sci., 38:133–136, 1985.
    :arg m: number of states
    :type m: integer
    :returns: a dfa
    :rtype: DFA"""

    if m < 3:
        raise TestsError("number of states must be greater than 2")
    f = DFA()
    f.setSigma(["a", "b", "c"])
    f.States = range(m)
    f.setInitial(0)
    f.addFinal(0)
    f.addTransition(0, "b", 1)
    f.addTransition(1, "b", 0)
    f.addTransition(0, "a", 1)
    f.addTransition(1, "a", 2)
    f.addTransition(0, "c", m - 1)
    f.addTransition(1, "c", 1)
    for i in range(2, m):
        f.addTransition(i, "a", (i + 1) % m)
        f.addTransition(i, "b", i)
        f.addTransition(i, "c", i)
    return f


def reversalternaryWC(m=5):
    """Worst case automata for reversal(DFA) ternary alphabet
       
        :arg m: number of states
        :type m: integer
        :returns: a dfa
        :rtype: DFA"""
    if m < 3:
        raise TestsError("number of states must be greater than 2")
    d = DFA()
    d.setSigma(["a", "b", "c"])
    d.setInitial(0)
    d.addFinal(0)
    d.States = range(m)
    d.addTransition(0, "a", m - 1)
    d.addTransition(0, "c", 0)
    d.addTransition(0, "b", 0)
    d.addTransition(1, "c", m - 1)
    d.addTransition(1, "b", 0)
    d.addTransition(1, "a", 0)
    for i in range(2, m):
        d.addTransition(i, "a", i - 1)
        d.addTransition(i, "c", i - 1)
        d.addTransition(i, "b", i)
    return d


def reversalbinaryWC(m=5):
    """Worst case automata for reversal(DFA) binary
    ..seealso:: G. Jir{\'a}skov{\'a} and J. S\v ebej. Note on Reversal of binary regular languages. Proc. DCFS 2011,
    LNCS 6808, Springer, pp 212-221.
    @arg m: number of states
    @type m: integer
    @returns: a dfa
    @rtype: DFA"""

    if m < 2:
        raise TestsError("number of states must be greater than 1")
    d = DFA()
    d.setSigma(["a", "b"])
    d.States = range(m)
    d.setInitial(0)
    d.addFinal(m - 1)
    d.addTransition(0, "a", 1)
    d.addTransition(0, "b", 0)
    d.addTransition(1, "b", 0)
    if m == 2:
        d.addTransition(1, "a", 0)
    else:
        d.addTransition(1, "a", 2)
        d.addTransition(2, "a", 0)
        if m == 3:
            d.addTransition(2, "b", 2)
        else:
            d.addTransition(2, "b", 3)
            d.addTransition(3, "b", 2)
            d.addTransition(3, "a", 4)
            d.addTransition(m - 1, "a", 3)
            d.addTransition(m - 1, "b", m - 1)
            for i in range(4, m - 1):
                d.addTransition(i, "a", i + 1)
                d.addTransition(i, "b", i)
    return d


def shuffleWC(m=3, n=3):
    """Worst case automata for shuffle(DFA,DFA) with m.n>1

    ..seealso::
     C. Campeanu, K. Salomaa, and S. Yu. Tight lower bound for
    the state complexity of shuffle of regular languages.
    Journal of Automata, Languages and Combinatorics, 7(3):303–310, 2002.

    :arg m: number of states
    :arg n: number of states
    :type m: integer
    :type n: integer
    :returns: two dfas 
    :rtype: (DFA, DFA)"""
    if n < 2 or m < 2:
        raise TestsError("number of states must be both greater than 1")
    d1, d2 = DFA(), DFA()
    d1.States = range(m)
    d1.setSigma(["a", "b", "c", "d", "f"])
    d1.setInitial(0)
    d1.addFinal(0)
    for i in range(m):
        d1.addTransition(i, "a", (i + 1) % m)
        if i != m - 1:
            d1.addTransition(i, "c", i + 1)
        d1.addTransition(i, "d", i)
        if i != 0:
            d1.addTransition(i, "f", i)
    d2.States = range(n)
    d2.setSigma(["a", "b", "c", "d", "f"])
    d2.setInitial(0)
    d2.addFinal(0)
    for i in range(n):
        d2.addTransition(i, "b", (i + 1) % n)
        d2.addTransition(i, "c", i)
        if i != n - 1:
            d2.addTransition(i, "d", i + 1)
        if i != 0:
            d2.addTransition(i, "f", i)
    return d1, d2


def starDisjWC(m=6, n=5):
    """Worst case automata for starDisj(DFA,DFA) with m.n>1

     ..seealso: Arto Salomaa, Kai Salomaa, and Sheng Yu. 'State complexity of
    combined operations'. Theor. Comput. Sci., 383(2-3):140–152, 2007.
    :arg m: number of states
    :arg n: number of states
    :type m: integer
    :type n: integer
    :returns: two dfas 
    :rtype: (DFA,DFA)"""

    if n < 2 or m < 2:
        raise TestsError("number of states must be both greater than 1")
    d1, d2 = DFA(), DFA()
    d1.States = range(m)
    d1.setSigma(["a", "b", "c"])
    d1.setInitial(0)
    d1.addFinal(0)
    for i in range(m):
        d1.addTransition(i, "a", (i + 1) % m)
        d1.addTransition(i, "b", i)
        if i != 0:
            d1.addTransition(i, "c", i)
    d1.addTransition(0, "c", 1)
    d2.States = range(n)
    d2.setSigma(["a", "b", "c"])
    d2.setInitial(0)
    d2.addFinal(0)
    for i in range(n):
        d2.addTransition(i, "b", (i + 1) % n)
        d2.addTransition(i, "a", i)
        if i != 0:
            d2.addTransition(i, "c", i)
    d2.addTransition(0, "c", 1)
    return d1, d2


def starInterBC(m=3, n=3):
    """Bad case automata for starInter(DFA,DFA) with m,n>1
    ..seealso:: Arto Salomaa, Kai Salomaa, and Sheng Yu. 'State complexity of
    combined operations'. Theor. Comput. Sci., 383(2-3):140–152, 2007.
    :arg m: number of states
    :arg n: number of states
    :type m: integer
    :type n: integer
    :returns: two dfas 
    :rtype: (DFA,DFA)"""

    if n < 2 or m < 2:
        raise TestsError("number of states must be both greater than 1")
    d1, d2 = DFA(), DFA()
    d1.setSigma(["a", "b", "c", "d", "e"])
    d1.States = range(m)
    d1.setInitial(0)
    d1.addFinal(m - 1)
    for i in range(m):
        d1.addTransition(i, "a", (i + 1) % m)
        d1.addTransition(i, "b", i)
        d1.addTransition(i, "c", i)
        d1.addTransition(i, "d", i)
        d1.addTransition(i, "e", i)
    d2.setSigma(["a", "b", "c", "d", "e"])
    d2.States = range(n)
    d2.setInitial(0)
    d2.addFinal(n - 1)
    for i in range(n):
        d2.addTransition(i, "b", (i + 1) % n)
        d2.addTransition(i, "a", i)
        d2.addTransition(i, "c", n - 2)
        if i == n - 2:
            d2.addTransition(i, "d", n - 1)
        elif i == n - 1:
            d2.addTransition(i, "d", n - 2)
        else:
            d2.addTransition(i, "d", i)
        if i > n - 4:
            d2.addTransition(i, "e", i)
        else:
            d2.addTransition(i, "e", i + 1)
    return d1, d2


def disjWStarWC(m=6, n=5):
    """
     ..seealso:: Yuan Gao and Sheng Yu. 'State complexity of union and intersection
  combined with star and reversal'. CoRR, abs/1006.3755, 2010.
  :arg m: number of states
  :arg n: number of states
  :type m: integer
  :type n: integer
  :returns: two dfas 
  :rtype: (DFA,DFA)"""

    if n < 3 or m < 3:
        raise TestsError("number of states must be greater than 2")
    f1 = DFA()
    f1.setSigma(["a", "b", "c"])
    f1.States = range(m)
    f1.setInitial(0)
    f1.addFinal(m - 1)
    f1.addTransition(0, "a", 1)
    f1.addTransition(0, "b", 0)
    f1.addTransition(0, "c", 0)
    for i in range(1, m):
        f1.addTransition(i, "a", (i + 1) % m)
        f1.addTransition(i, "b", (i + 1) % m)
        f1.addTransition(i, "c", i)
    f2 = DFA()
    f2.setSigma(["a", "b", "c"])
    f2.States = range(n)
    f2.setInitial(0)
    f2.addFinal(n - 1)
    for i in range(n):
        f2.addTransition(i, "a", i)
        f2.addTransition(i, "b", i)
        f2.addTransition(i, "c", (i + 1) % n)
    return f1, f2


### worst cases for transition complexity


#######     UNION   ######

def unionWCTk2(m=6, n=6):
    """ @ worst-case family union where
    @m>=2 and n>=2 and k=2
    ..seealso:: Gao, Y., Salomaa, K., Yu, S.: Transition complexity of
    incomplete dfas. Fundam. Inform.  110(1-4), 143–158 (2011)
    @ the conjecture in this article fails for this family
    :arg m: number of states
    :arg n: number of states
    :type m: integer
    :type n: integer
    :returns: two dfas 
    :rtype: (DFA,DFA)"""

    if n < 2 or m < 2:
        raise TestsError("number of states must both  greater than 1")
    d1, d2 = DFA(), DFA()
    d1.setSigma(["a", "b"])
    d1.States = range(m)
    d1.setInitial(0)
    d1.addFinal(0)
    d1.addTransition(m - 1, "a", 0)
    for i in range(0, m - 1):
        d1.addTransition(i, "b", i + 1)
    d2.setSigma(["a", "b"])
    d2.States = range(n)
    d2.setInitial(0)
    d2.addFinal(n - 1)
    d2.addTransition(n - 1, "b", n - 1)
    for i in range(0, n - 1):
        d2.addTransition(i, "a", i + 1)
        d2.addTransition(i, "b", i)
    return d1, d2


def unionWCT2(n=6):
    """ @ worst-case family union where
    @m=1 and n>=2 and k=3
    @ Note that the same happens to m>=2 and n=1
    :arg n: number of states
    :type n: integer
    :returns: two dfas 
    :rtype: (DFA,DFA)"""
    m = 1
    if n < 2:
        raise TestsError("number of states must both  greater than 1")
    d1, d2 = DFA(), DFA()
    d1.setSigma(["a", "b", "c"])
    d1.States = range(m)
    d1.setInitial(0)
    d1.addFinal(0)
    d1.addTransition(0, "b", 0)
    d1.addTransition(0, "c", 0)

    d2.setSigma(["a", "b", "c"])
    d2.States = range(n)
    d2.setInitial(0)
    d2.addFinal(n - 1)
    d2.addTransition(0, "a", 0)
    d2.addTransition(0, "b", 1)
    for i in range(1, n):
        d2.addTransition(i, "b", (i + 1) % n)
        d2.addTransition(i, "a", i)
        d2.addTransition(i, "c", 1)
    return d1, d2


def unionWCT(m=6, n=6):
    """ @ worst-case family union where
    @m>=2 and n>=2 and k=3
    :arg m: number of states
    :arg n: number of states
    :type m: integer
    :type n: integer
    :returns: two dfas 
    :rtype: (DFA,DFA)"""

    if n < 2 or m < 2:
        raise TestsError("number of states must both  greater than 1")
    d1, d2 = DFA(), DFA()
    d1.setSigma(["a", "b", "c"])
    d1.States = range(m)
    d1.setInitial(0)
    d1.addFinal(m - 1)
    d1.addTransition(0, "a", 1)
    d1.addTransition(0, "c", 0)
    for i in range(1, m):
        d1.addTransition(i, "a", (i + 1) % m)
        d1.addTransition(i, "b", 0)
        d1.addTransition(i, "c", i)
    d2.setSigma(["a", "b", "c"])
    d2.States = range(n)
    d2.setInitial(0)
    d2.addFinal(n - 1)
    d2.addTransition(0, "a", 0)
    d2.addTransition(0, "b", 1)
    for i in range(1, n):
        d2.addTransition(i, "b", (i + 1) % n)
        d2.addTransition(i, "a", i)
        d2.addTransition(i, "c", 1)
    return d1, d2


### CONCAT 

def concatWCT2(n=6):
    """ @ worst-case family concatenation where
    @m=1 and n>=2 and k=3
    :arg n: number of states
    :type n: integer
    :returns: two dfas 
    :rtype: (DFA,DFA)"""
    m = 1
    if n < 2:
        raise TestsError("number of states must both  greater than 1")
    d1, d2 = DFA(), DFA()
    d1.setSigma(["a", "b", "c"])
    d1.States = range(m)
    d1.setInitial(0)
    d1.addFinal(0)
    d1.addTransition(0, "b", 0)
    d1.addTransition(0, "c", 0)

    d2.setSigma(["a", "b", "c"])
    d2.States = range(n)
    d2.setInitial(0)
    d2.addFinal(n - 1)
    d2.addTransition(0, "a", 0)
    d2.addTransition(0, "b", 1)
    for i in range(1, n):
        d2.addTransition(i, "b", (i + 1) % n)
        d2.addTransition(i, "a", i)
        d2.addTransition(i, "c", (i + 1) % n)
    return d1, d2


def concatWCT3(m=6):
    """ @ worst-case family concatenation where
    @m>=2 and n=1 and k=3
    :arg m: number of states
    :type m: integer
    :returns: two dfas 
    :rtype: (DFA,DFA)"""
    n = 1
    if m < 2:
        raise TestsError("number of states must both  greater than 1")
    d1, d2 = DFA(), DFA()
    d1.setSigma(["a", "b", "c"])
    d1.States = range(m)
    d1.setInitial(0)
    d1.addFinal(m - 1)
    d1.addTransition(0, "a", 0)
    d1.addTransition(0, "b", 1)
    d1.addTransition(0, "c", 1)
    d1.addTransition(1, "a", 1)
    d1.addTransition(1, "b", 2)
    for i in range(2, m):
        d1.addTransition(i, "b", (i + 1) % m)
        d1.addTransition(i, "c", (i + 1) % m)
        d1.addTransition(i, "a", i)
    d2.setSigma(["a", "b", "c"])
    d2.States = range(n)
    d2.setInitial(0)
    d2.addFinal(0)
    d2.addTransition(0, "c", 0)
    d2.addTransition(0, "b", 0)

    return d1, d2


def concatWCT(m=6, n=6):
    """ @ worst-case family concatenation where
    @m>=2 and n>=2 and k=3
    :arg m: number of states
    :arg n: number of states
    :type m: integer
    :type n: integer
    :returns: two dfas 
    :rtype: (DFA,DFA)"""

    if n < 2 or m < 2:
        raise TestsError("number of states must both  greater than 1")
    d1, d2 = DFA(), DFA()
    d1.setSigma(["a", "b", "c"])
    d1.States = range(m)
    d1.setInitial(0)
    d1.addFinal(m - 1)
    d1.addTransition(0, "a", 1)
    d1.addTransition(0, "c", 0)
    for i in range(1, m):
        d1.addTransition(i, "a", (i + 1) % m)
        d1.addTransition(i, "b", 0)
        d1.addTransition(i, "c", i)
    d2.setSigma(["a", "b", "c"])
    d2.States = range(n)
    d2.setInitial(0)
    d2.addFinal(n - 1)
    d2.addTransition(0, "a", 0)
    d2.addTransition(0, "b", 1)
    for i in range(1, n):
        d2.addTransition(i, "b", (i + 1) % n)
        d2.addTransition(i, "a", i)
        d2.addTransition(i, "c", 1)
    return d1, d2


#####  Star

def starWCT(m=5):
    """ @ worst-case family star where
    @m>=2 and k=2
    :arg m: number of states
    :type m: integer
    :returns: dfa 
    :rtype: DFA"""
    if m < 3:
        raise TestsError("number of states must be greater than 2")
    f = DFA()
    f.setSigma(["a", "b"])
    f.States = range(m)
    f.setInitial(0)
    f.addFinal(m - 1)
    f.addTransition(0, "a", 1)
    for i in range(1, m):
        f.addTransition(i, "a", (i + 1) % m)
        f.addTransition(i, "b", (i + 1) % m)
    return f


def starWCT1(m=5):
    """ @ worst-case family star where
    @m>=2 and k=2
    :arg m: number of states
    :type m: integer
    :returns: dfa 
    :rtype: DFA"""
    if m < 3:
        raise TestsError("number of states must be greater than 2")
    f = DFA()
    f.setSigma(["a", "b"])
    f.States = range(m)
    f.setInitial(0)
    f.addFinal(m - 1)
    f.addTransition(0, "b", 0)
    f.addTransition(0, "a", 1)
    f.addTransition(m - 2, "a", m - 1)
    f.addTransition(m - 1, "a", 0)
    for i in range(1, m - 2):
        f.addTransition(i, "a", (i + 1) % m)
        f.addTransition(i, "b", (i + 1) % m)
    return f


def universal(n, l= None, Finals=None, dialect=False, d=None):
    """Universal witness for state compelxity
    :arg int n: number of states
    :arg [str] l: alphabet
    :arg [int] Finals: list of final states
    :arg bool dialect: is it a dialect
    :returns: dfa 
    :rtype: DFA
    """
    if n < 3:
        raise TestsError("number of states must be greater than 2")
    u = DFA()
    u.States = range(n)
    if l is None:
        l = ("a", "b", "c")
    u.setSigma(list(l))
    u.setInitial(0)
    u.addFinal(n - 1)
    u.addTransition(0, "b", 1)
    u.addTransition(1, "b", 0)
    if "c" in l:
        u.addTransition(n - 1, "c", 0)
    for i in range(n):
        u.addTransition(i, "a", (i + 1) % n)
        if i >= 2:
            u.addTransition(i, "b", i)
        if i != n - 1 and "c" in l:
            u.addTransition(i, "c", i)
    return u


def nCr(n, r):
    import math

    if r > n:
        return 0
    else:
        f = math.factorial
        return f(n) / (f(r) * f(n - r))


def boundarySC(n, k):
    return 4 ** (n - 1) - nCr(n - 1, k - 1) + 2 ** (n - k) * 2 ** (n - 1) - 3 ** (n - k) * 2 ** (k - 1) + 2 ** (
        k - 1) * 2 ** (n - 1) - 3 ** (k - 1) * 2 ** (n - k) + 1

# Don't care automata

def dcMilano1(n):
    """Return the special dcNFA to prove the titness of proposed bound

    .. versionadded:: 0.9.8

    :param n: number of 'columns'
    :type n: int
    :rtype: NFA"""
    new = fa.NFA()
    st = []
    for _ in xrange(3 * n):
        s = new.addState()
        st.append(s)
    new.setInitial([st[n - 1]])
    for s in xrange(n):
        new.addTransition(st[n - 1], 'c', st[s])
    for s in xrange(3):
        for r in xrange(n - 1):
            new.addTransition(st[(n * s) + r], 'a', st[(n * s) + r + 1])
        new.addTransition(st[(n * s) + n - 1], 'a', st[0])
    for s in xrange(n):
        new.addTransition(st[s], 'b', st[s + n])
        new.addTransition(st[s + n], 'b', st[s + 2 * n])
        new.addTransition(st[s + 2 * n], 'b', st[s])
    new.addFinal(st[n - 1])
    return new


def dcMilano2(n):
    """Return the special dcNFA to prove the titness of proposed bound

    .. versionadded:: 0.9.8

    :param n: number of 'columns'
    :type n: int
    :rtype: NFA"""
    new = fa.NFA()
    st = []
    for _ in xrange(3 * n):
        s = new.addState()
        st.append(s)
    for s in xrange(n):
        new.addInitial(st[s])
    for s in xrange(3):
        for r in xrange(n - 1):
            new.addTransition(st[(n * s) + r], 'a', st[(n * s) + r + 1])
        new.addTransition(st[(n * s) + n - 1], 'a', st[0])
    for s in xrange(n):
        new.addTransition(st[s], 'b', st[s + n])
        new.addTransition(st[s + n], 'b', st[s + 2 * n])
        new.addTransition(st[s + 2 * n], 'b', st[s])
    new.addFinal(st[n - 1])
    return new

# Closure operations

def suffWCe(m=3):
    """Witness for suff(L) when L does not have empty as a quotient

     :rtype: DFA

     ..seealso:
          Janusz A. Brzozowski, Galina Jirásková, Chenglong Zou,
          Quotient Complexity of Closed Languages.
          Theory Comput. Syst. 54(2): 277-292 (2014)

     """
    if m< 3:
        raise TestsError("number of states must be greater than 2")
    f = DFA()
    f.setSigma(["a", "b"])
    f.States = range(m)
    f.setInitial(0)
    f.addFinal(0)
    f.addTransition(0, "a", 1)
    f.addTransition(1, "a", 2)
    f.addTransition(0, "b", 0)
    f.addTransition(1, "b", 0)
    for i in range(2, m):
        f.addTransition(i, "a", (i + 1) % m)
        f.addTransition(i, "b", i)
    return f

def suffWCd(m=3):

    """Witness for suff(L) when L has  empty as a quotient

    :rtype: DFA

    ..seealso: as above
    """
    if m< 3:
        raise TestsError("number of states must be greater than 2")
    f = DFA()
    f.setSigma(["a", "b"])
    f.States = range(m)
    f.setInitial(0)
    f.addFinal(0)
    f.addTransition(0, "a", 1)
    f.addTransition(1, "a", 2)
    f.addTransition(0, "b", m-1)
    f.addTransition(1, "b", 0)
    f.addTransition(m-1, "b", m-1)
    f.addTransition(m-1, "a", m-1)
    for i in range(2, m-1):
        f.addTransition(i, "a", (i + 1) % (m-1))
        f.addTransition(i, "b", i)
    return f

def suffWCsynt(m=3):
  """ Worst case witness for synt of suff(L)

  """
  if m< 3:
     raise TestsError("number of states must be greater than 2")
  f = DFA()
  f.setSigma(["a", "b", "c", "d", "e"])
  f.States = range(m)
  f.setInitial(0)
  f.addFinal(m-1)
  f.addTransition(0, "a", 0)
  f.addTransition(0, "b", 0)
  f.addTransition(0, "c", 0)
  f.addTransition(0, "d", 0)
  f.addTransition(0, "e", 1)
  f.addTransition(1, "a", 2)
  f.addTransition(1, "b", 2)
  f.addTransition(1, "c", 1)
  f.addTransition(1, "d", 1)
  f.addTransition(1, "e", 1)
  f.addTransition(2, "b", 1)
  f.addTransition(2, "e", 1)
  f.addTransition(2, "c", 2)
  f.addTransition(2, "d", 2)
  f.addTransition(2, "a", 3)
  for i in range(3,m-1):
    f.addTransition(i, "a", (i+1)% m)
    f.addTransition(i,"b",i)
    f.addTransition(i,"c",i)
    f.addTransition(i,"d",i)
    f.addTransition(i,"e",1)
  f.addTransition(m-1,"a",1)
  f.addTransition(m-1,"c",1)
  f.addTransition(m-1,"e",1)
  f.addTransition(m-1,"d",0)
  return f

def booleanWCSymGrp(m=3):
  """Witness for symmetric group

   :rtype: DFA
   ..seealso:
   	Jason Bell, Janusz A. Brzozowski, Nelma Moreira, Rogério Reis.
   	Symmetric Groups and Quotient Complexity of Boolean Operations.
   	ICALP (2) 2014: 1-12
  """
  if m< 3:
     raise TestsError("number of states must be greater than 2")
  f = DFA()
  f.setSigma(["a", "b"])
  f.States = range(m)
  f.setInitial(0)
  f.addFinal(0)
  f.addFinal(1)
  f.addTransition(0, "a", 1)
  f.addTransition(1, "a", 0)
  f.addTransition(0, "b", 1)
  f.addTransition(1, "b", 2)
  for i in range(2, m):
    f.addTransition(i, "b", (i + 1) % m)
    f.addTransition(i, "a", i)
  return f

def suffFreeSyntWC(m=5):
    """

    """
    if m < 5:
     raise TestsError("number of states must be greater than 2")
    f = DFA()
    f.setSigma(["a", "b", "c","d", "e"])
    f.States = range(m)
    f.setInitial(0)
    f.addFinal(1)
    f.addTransition(0,"a",m-1)
    f.addTransition(0,'b',m-1)
    f.addTransition(0,'c',m-1)
    f.addTransition(0,'d',m-1)
    f.addTransition(0,'e',1)
    f.addTransition(1,"a",2)
    f.addTransition(1,'c',1)
    f.addTransition(1,'e',m-1)
    f.addTransition(1,'d',m-1)
    f.addTransition(1,'b',2)
    f.addTransition(2,'b',1)
    f.addTransition(2,"a",3)
    f.addTransition(2,"c",2)
    f.addTransition(2,"e",m-1)
    f.addTransition(2,'d',2)
    f.addTransition(1,'d',m-1)
    f.addTransition(m-2,'c',1)
    f.addTransition(m-2,'a',1)
    for sym in f.Sigma:
        f.addTransition(m-1,sym,m-1)
    for i in range(3,m-1):
        f.addTransition(i,"b",i)
        if i != m-2:
            f.addTransition(i,"c",i)
            f.addTransition(i,"a",(i+1))
        f.addTransition(i,"d",i)
        f.addTransition(i,"e",m-1)
    return f




#shuffle

def shuffle_ReBC(n=3):
    """ Withess for Shuffle toDFA
    """

    if n < 2:
        raise TestsError("at least n=2")
    rs = star(concat(atom('a0'), atom('b0')))
    for i in range(1, n):
        rs = shuffle(rs,star(concat(atom('a'+str(i)),atom('b'+str(i)))))
    return rs

def shuffle_ReBCF(n=3):
    """
    :param n:
    :return:
    """
    if n<2:
        raise TestsError("at least n=2")
    rs = concat(atom('a0'),atom('b0'))
    for i in range(1,n):
        rs = shuffle(rs,concat(atom('a'+str(i)),atom('b'+str(i))))
    return rs

def shuffle_NFABC(n=3):
    """

    :param  int n:
    :rtype: atom
    """
    if n<2:
        raise TestsError("at least n=2")
    ns = star(concat(atom('a0'), atom('b0'))).toNFA()
    for i in range(1, n):
        ni = star(concat(atom('a'+str(i)), atom('b'+str(i)))).toNFA()
        ns = ns.shuffle(ni)
    return ns


def shuffle_ReWC(n=4):
    """

    :param int n:
    :rtype: atom
    """
    if n<2:
        raise TestsError("at least n=2")
    rs = atom('a0')
    for i in range(1, n):
        rs = shuffle(rs, atom('a'+str(i)))
    return rs


def intersectionReWC(n, sym = 'a'):
    if n == 0:
        return epsilon()
    if n == 1:
        return star(atom(sym))
    else:
        s1 = set()
        s2 = set()
        for i in range(1,n+1):
            ri = str2xre(sym * i)
            s1.add(star(ri))
            s2.add(ri)
        r1 = sdisj(frozenset(s1))
        r2 = sstar(sdisj(frozenset(s2)))
        return sconj(frozenset([r1,r2]))


def revFibonnacci(n):
    a = DFA ()
    a.setSigma (["a", "b"]),
    for i in range (n):
        a.addState (i)
    a.setInitial (0)
    for i in range (n):
        if i % 2 != 0:
            if i < n - 1:
                a.addTransition (i, 'b', i + 1)
            if i < n - 2:
                a.addTransition (i, 'a', i + 2)
        else:
            if i < n - 2:
                a.addTransition (i, 'b', i + 2)
            if i < n - 1:
                a.addTransition (i, 'a', i + 1)
    a.addFinal (n - 1)
    return a


# -*- coding: utf-8 -*-
"""Code theory module

.. versionadded:: 1.0

.. *Authors:* Rogério Reis, Nelma Moreira & Stavros Konstantinidis

.. *This is part of FAdo project*   http://fado.dcc.fc.up.pt.

.. *Copyright:* 1999-2014 Rogério Reis & Nelma Moreira {rvr,nam}@dcc.fc.up.pt

.. This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   675 Mass Ave, Cambridge, MA 02139, USA."""
import copy

import reex
from common import *
import fio
from transducers import SFT, infixTransducer, prefixTransducer, suffixTransducer, outfixTransducer, hypercodeTransducer
import fa
import fl


def fixedHierSubset(x, y):
    """ Returns whether x==y, or the fixed property with name x is a subset of y
    Currently (Jan 2015) the fixed properties names are   'UD_codes',
    'Prefix_codes', 'Suffix_codes', 'Infix_codes', 'Outfix_codes', 'Hypercodes'

    :param tuple x: first argument
    :param tuple y: second argument
    :rtype: bool"""
    if x == y:
        return True
    if x[0] != 'Fixed' or y[0] != 'Fixed':
        return False
    if x[1] == 'Hypercodes' or y[1] == 'UD_codes':
        return True
    if x[1] == 'UD_codes' or y[1] == 'Hypercodes':
        return False
    if x[1] == 'Infix_codes':
        if y[1] == 'Outfix_codes':
            return False
        else:
            return True
    if x[1] == 'Outfix_codes':
        if y[1] == 'Infix_codes':
            return False
        else:
            return True
    if y[1] == 'Infix_codes' or y[1] == 'Outfix_codes':
        return False
    if x[1] == y[1]:
        return True
    else:
        return False


def isSubclass(p1, p2):
    """Which property (language class) is a subclass of the other (if any).
       It returns 1 if p1 is a subclass of p2; 2 if p2 is a subclass of p1;
       3 if they are equal; 0 otherwise

    :param IPTProp p1: an input preserving transducer property
    :param IPTProp p2: an input preserving transducer property
    :rtype: int"""

    def _lessEqualID(first, second):
        """ Returns whether the property with ID X is a subset of that with ID Y.
        This is the case when X has a subset X1, say, such that each property named
        in X1 is  a subset of some property named in Y
        :param set first: first argument
        :param set second: second argument
        :rtype: bool"""
        yl = [y for y in second]
        ydel = []
        for x in first:
            for y in yl:
                if fixedHierSubset(x, y):
                    ydel.append(y)
            yl = [y for y in yl if y not in ydel]
            if len(yl) == 0:
                return True
        return False

    if p1.Sigma != p2.Sigma:
        return 0
    if p1.ID <= p2.ID:
        if p2.ID <= p1.ID:
            return 3
        else:
            return 2
    if p2.ID <= p1.ID:
        return 1
    if _lessEqualID(p1.ID, p2.ID):
        return 1
    if _lessEqualID(p2.ID, p1.ID):
        return 2
    return 0


def unionOfIDs(first, second):
    """ Returns the "union" of the two sets: property ID X and propoerty ID Y.
    The result is the set union minus any element in one set that names
    a property containing a property named in the other set

    :param set first: first argument
    :param set second: second argument
    :rtype: set"""
    z = set() | first
    xl = [x for x in first]
    yl = [y for y in second]
    while len(xl) > 0 and len(yl) > 0:
        x = xl.pop()
        ydel = []
        for j in range(0, len(yl)):
            y = yl[j]
            if fixedHierSubset(x, y):
                ydel.append(y)
            elif fixedHierSubset(y, x):
                z.add(y)
                ydel.append(y)
                z.remove(x)
                break
            else:
                pass
        yl = [y for y in yl if y not in ydel]
    return z | set(yl)


class CodeProperty(object):
    """
    .. seealso:: K. Dudzinski and S. Konstantinidis: Formal descriptions of code properties: decidability, complexity,
       implementation. International Journal of Foundations of Computer Science 23:1 (2012), 67--85.

    :var Sigma: the alphabet"""
    def __init__(self, name, alph):
        if name is None:
            raise CodingTheoryError('new property must have a name')
        self.Sigma = alph
        self.ID = set()
        if type(name) is set:
            self.ID.update(name)
        else:
            self.ID.add(name)

    @abstractmethod
    def satisfiesP(self, aut):
        """Satisfaction of the property by the automaton language

        :param NFA|DFA aut: the automaton
        :rtype: bool"""
        pass

    @abstractmethod
    def notSatisfiesW(self, aut):
        """Return a witness of non-satisfaction of the property by the automaton language

        :param DFA|NFA aut: the automaton
        :return: word witness tuple
        :rtype: tuple"""
        pass

    @abstractmethod
    def maximalP(self, aut, U=None):
        """Tests if the language is maximal w.r.t. the property

        :param DFA|NFA U: Universe of permitted words (Sigma^* as default)
        :param DFA|NFA aut: the automaton
        :rtype: bool"""
        pass

    @abstractmethod
    def notMaximalW(self, aut, U=None):
        """Witness of non maximality

        :param DFA|NFA aut: the automaton
        :param DFA|NFA U: Universe of permitted words (Sigma^* as default)
        :return: a witness
        :rtype: str"""
        pass


def pickFrom(s, ell):
    """Returns a random string of length ell over {0,1,...,s-1}

    :param int s: the size of the alphabet (should be <= 10)
    :param int ell: the length of the desired string
    :return: a string
    :rtype: str"""
    random.seed()
    w = ""
    last_digit = s-1
    for _ in range(ell):
        d = random.randint(0,last_digit)
        w += str(d)
    return w


class IPTProp(CodeProperty):
    """Input Preserving Transducer Property

    .. inheritance-diagram:: IPTProp

    :var SFT Aut: the transducer defining the property
    :var set Sigma: alphabet"""
    def __init__(self, aut, name=None):
        """ Constructor
        :param SFT aut: Input preserving transducer"""
        if name is None:
            name = aut.codeOfTransducer()
        super(IPTProp, self).__init__(name, aut.Sigma)
        self.Aut = aut

    def __eq__(self, other):
        return isinstance(other, IPTProp) and (isSubclass(self, other)==3)

    def __ne__(self, other):
        return not self == other

    def __and__(self, other):
        """ Conjunction of two properties

        :param IPTProp other: right hand operand
        :rtype: IPTProp """
        sub = isSubclass(self, other)
        if sub == 1 or sub == 3:
            return self
        if isinstance(other, IATProp):
            if sub == 2:
                return other
            nt = other.Aut.dup()
            n = nt.addState()
            nt.addInitial(n)
            nt.addFinal(n)
            for s in nt.Sigma:
                nt.addTransition(n, s, s, n)
            name = unionOfIDs(self.ID, other.ID)
            pty = IPTProp(self.Aut | nt, name)
            return pty
        else:
            if sub == 2:
                return other
            name = unionOfIDs(self.ID, other.ID)
            pty = IPTProp(self.Aut | other.Aut, name)
            return pty

    def notSatisfiesW(self, aut):
        """Return a witness of non-satisfaction of the property by the automaton language

        :param DFA|NFA aut: the automaton
        :return: word witness pair
        :rtype: tuple"""
        u, v, w = self.Aut.inIntersection(aut).outIntersection(aut).nonFunctionalW()
        if u == v:
            return u, w
        else:
            return u, v

    def satisfiesP(self, aut):
        """Satisfaction of the property by the automaton language

        :param DFA|NFA aut: the automaton
        :rtype: bool"""
        return self.notSatisfiesW(aut) == (None, None)

    def notMaximalW(self, aut, U=None):
        """Tests if the language is maximal w.r.t. the property

        :param DFA|NFA aut: the automaton
        :param DFA|NFA U: Universe of permitted words (Sigma^* as default)
        :rtype: bool
        :raises PropertyNotSatisfied: if not satisfied"""
        if not self.satisfiesP(aut):
            raise PropertyNotSatisfied("Property is not satisfied")
        l = (self.Aut.runOnNFA(aut) | self.Aut.inverse().runOnNFA(aut) | aut).__invert__()
        if U:
            l = l & U.toNFA()
        try:
            w = l.witness()
        except DFAEmptyDFA:
            w = None
        return w

    def maximalP(self, aut, U=None):
        """Tests if the language is maximal w.r.t. the property

        :param NFA aut: the automaton
        :param NFA U: Universe of permitted words (Sigma^* as default)
        :rtype: bool"""
        return self.notMaximalW(aut.toNFA(), U) is None


    def notMaxStatW(self, aut, ell, n=2000, ov_free=False):
        """Returns a word of length ell to add into aut or None; simpler version of
        function nonMaxStatFEpsW

        :param NFA aut: the automaton
        :param int ell: the length of the words in aut
        :param int n: number of words to try
        :return: a string or None
        :rtype: str"""
        #original
        #if hasattr(self, "AutDetect"):
        # b = self.AutDetect.runOnNFA(aut).trim()
        #else:
        #  b = (self.Aut | self.Aut.inverse()).runOnNFA(aut).trim()
        b = (self.Aut | self.Aut.inverse()).inIntersectionSlow(aut).toOutNFA().trim()
        s = len(aut.Sigma)
        # can use ell = len(aut.witness()),  if ell was not given
        tr = 1
        while tr <= n:
            w = pickFrom(s, ell)
            if ov_free and not overlapFreeP(w):
                continue
            if not b.evalWordP(w): # == False:
                return w
            tr += 1
        return None    # in this case, tr == n+1


    def makeCode(self, N, ell, s, n=2000,ov_free=False):
        """Returns an NFA and a list W of up to N words of length ell, such that the NFA
        accepts W, which is an error-detecting language. The alphabet to use is
        {0,1,...,s-1}. where s <= 10.

        :param int N: the number of words to construct
        :param int ell: the codeword length
        :param int s: the alphabet size (must be <= 10)
        :param int n: number of tries when needing a new word
        :return: an automaton and a list of strings
        :rtype: tuple"""
        w = pickFrom(s, ell)
        if ov_free:
            while not overlapFreeP(w):
                w = pickFrom(s, ell)
        W = [w]
        aut = fl.FL(W).toNFA()
        cnt = 1
        self.AutDetect = (self.Aut | self.Aut.inverse())
        while cnt < N:
            w = self.notMaxStatW(aut, ell, n,ov_free)
            if w is None: break
            W.append(w)
            aut = fl.FL(W).toNFA()
            cnt += 1
        return aut, W

    def makeCodeO(self, N, ell, s, n=2000, end=None, ov_free=False):
        """Returns an NFA and a list W of up to N words of length ell, such that the NFA
        accepts W, which is an error-detecting language. The alphabet to use is
        {0,1,...,s-1}. where s <= 10.

        :param int N: the number of words to construct
        :param int ell: the codeword length
        :param int s: the alphabet size (must be <= 10)
        :param int n: number of tries when needing a new word
        :param Word end: a Word or None that should much the end of code words
        :param Boolean ov_free: if True code words much be overlap free
        :return: an automaton and a list of strings
        :rtype: tuple

        Note: not ov_free and end defined simultaneously
        Note: end should be a Word
        """


        a = fl.sigmaInitialSegment(set([str(i) for i in range(s)]), ell, True)
        g = fl.RndWGen(a)
        w = g.next()
        if ov_free and end is not None:
            l_end = len(end)
            end = list(end)
            we = w[-l_end:]
            while we != end or not overlapFreeP(w):
                w = g.next()
                we = w[-l_end:]
        if ov_free and end is None:
            while not overlapFreeP(w):
                w = g.next()
        if not ov_free and end is not None:
            l_end= len(end)
            end = list(end)
            we = w[-l_end:]
            while we != end:
                w = g.next()
                we = w[-l_end:]
        W = [w]
        aut = fl.FL(W).MADFA()#toNFA()
        cnt = 1
        self.AutDetect = (self.Aut | self.Aut.inverse())
        while cnt < N:
            b = self.AutDetect.runOnNFA(aut).trim()
            tr = 1
            while tr <= n:
                w = g.next()
                if ov_free:
                    if not overlapFreeP(w):
                        continue
                if end is not None:
                    if  w[-l_end:] != end:
                        continue
                if not b.evalWordP(w):  # == False:
                    W.append(w)
                    aut = fl.FL(W).MADFA()#toNFA()
                    cnt += 1
                    break
                tr += 1
            if tr > n: break
        return aut, W


    def addToCode(self, aut, N, n=2000):
        """Returns an NFA and a list W of up to N words of length ell, such that the NFA
        accepts L(aut) union W, which is an error-detecting language.
        ell is computed from aut

        :param NFA aut: the automaton
        :param int N: the number of words to construct
        :param int n: number of tries when needing a new word
        :return: an automaton and a list of strings
        :rtype: tuple"""
        a_new = aut.dup()
        W = []
        u = aut.witness()
        if u is None:
            return a_new, W
        ell = len(u)
        cnt = 0
        while cnt < N:
            w = self.notMaxStatW(a_new, ell, n)
            if w is None: break
            else:
                W.append(w)
                a_new = a_new | fl.FL([w]).trieFA().toNFA()
                cnt += 1
        return a_new, W

ErrDetectProp = IPTProp  # another name for IPTProp


class IATProp(IPTProp):
    """Input Altering Transducer Property

    .. inheritance-diagram:: IATProp"""

    def __and__(self, other):
        """ Conjunction of two properties

        :param IPTProp other: right hand operand
        :rtype: IPTProp """
        sub = isSubclass(self, other)
        if sub == 1 or sub == 3:
            return self
        if not isinstance(other, IATProp):
            if sub == 2:
                return other
            nt = self.Aut.dup()
            n = nt.addState()
            nt.addInitial(n)
            nt.addFinal(n)
            for s in nt.Sigma:
                nt.addTransition(n, s, s, n)
            name = unionOfIDs(self.ID, other.ID)
            pty = IPTProp(other.Aut | nt, name)
            return pty
        else:
            if sub == 2:
                return other
            name = unionOfIDs(self.ID, other.ID)
            pty = IATProp(self.Aut | other.Aut, name)
            return pty

    def notSatisfiesW(self, aut):
        """Return a witness of non-satisfaction of the property by the automaton language

        :param DFA|NFA aut: the automaton
        :return: word witness pair
        :rtype: tuple"""
        return self.Aut.inIntersection(aut).outIntersection(aut).nonEmptyW()


class TrajProp(IATProp):
    """Class of trajectoty properties

    .. inheritance-diagram:: TrajProp"""

    def __init__(self, aut, Sigma):
        """Constructor

        :param DFA|NFA aut: regular expression over {0,1}
        :param set Sigma: the alphabet"""
        super(TrajProp, self).__init__(self.trajToTransducer(aut, Sigma))

    @staticmethod
    def trajToTransducer(traj, Sigma):
        """ Input Altering Tranducer corresponding to a Trajectory

        :param NFA traj: trajectory language
        :param set Sigma: alphabet
        :rtype: SFT"""
        t = SFT()
        t.Sigma = Sigma
        for s, sn in enumerate(traj.States):
            no = t.stateIndex((sn, False), True)
            yes = t.stateIndex((sn, True), True)
            if s in traj.Final:
                t.addFinal(yes)
            if s in traj.Initial:
                t.addInitial(no)
            for b in traj.delta.get(s, {}):
                for c in Sigma:
                    for st in traj.delta[s][b]:
                        fNo = t.stateIndex((traj.States[st], False), True)
                        fYes = t.stateIndex((traj.States[st], True), True)
                        if b == "0":
                            t.addTransition(no, c, c, fNo)
                            t.addTransition(yes, c, c, fYes)
                        if b == "1":
                            t.addTransition(no, Epsilon, c, fYes)
                            t.addTransition(yes, Epsilon, c, fYes)
        return t


class FixedProp(CodeProperty):
    """ Abstract class for fixed properties """

    @abstractmethod
    def notSatisfiesW(self, aut):
        """Test whether the language is a code.

           :param DFA|NFA aut: the automaton
           :return: two different factorizations of the same word
           :rtype: tuple of list"""
        pass

    @abstractmethod
    def satisfiesP(self, aut):
        """Satisfaction of the property by the automaton language

        :param NFA|DFA aut: the automaton
        :rtype: bool"""
        pass

    @abstractmethod
    def maximalP(self, aut, U=None):
        """Tests if the language is maximal w.r.t. the property

        :param DFA|NFA U: Universe of permitted words (Sigma^* as default)
        :param DFA|NFA aut: the automaton
        :rtype: bool"""
        pass

    @abstractmethod
    def notMaximalW(self, aut, U=None):
        """Witness of non maximality

        :param DFA|NFA aut: the automaton
        :param DFA|NFA U: Universe of permitted words (Sigma^* as default)
        :return: a witness
        :rtype: str"""
        pass


class UDCodeProp(FixedProp):
    """ Uniquely decodable Code Property

    .. inheritance-diagram:: UDCodeProp"""

    def __init__(self, alphabet):
        super(UDCodeProp, self).__init__(('Fixed', 'UD_codes'), alphabet)

    def notSatisfiesW(self, aut):
        """Test whether the language is a code.

        :param DFA|NFA aut: the automaton
        :return: two different factorizations of the same word
        :rtype: tuple of list"""
        def _findFactors(w, m):
            l = []
            i1 = 0
            while True:
                j = m.find('1', i1)
                if j < 0:
                    break
                l.append(w[i1:j + 1])
                i1 = j + 1
            return l

        n = aut.toNFA()   # .eliminateEpsilonTransitions()
        for i in n.Initial:
            if i in n.Final:
                return [Epsilon], [Epsilon, Epsilon]
        t = SFT()
        t.States = copy.deepcopy(n.States)
        t.setInitial(n.Initial)
        t.setFinal(n.Initial)
        t.setSigma(n.Sigma)
        t.setOutput(['1', '0'])
        for si in n.delta:
            for sym in n.delta[si]:
                for so in n.delta[si][sym]:
                    if so not in n.Final:
                        t.addTransition(si, sym, '0', so)
                    else:
                        t.addTransition(si, sym, '0', so)
                        for ss in t.Initial:
                            t.addTransition(si, sym, '1', ss)
        t.trim()
        foo = t.nonFunctionalW()
        if foo == (None, None, None):
            return None, None
        else:
            return _findFactors(foo[0], foo[1]), _findFactors(foo[0], foo[2])

    def maximalP(self, aut, U=None):
        """Tests if the language is maximal w.r.t. the property
        :param DFA|NFA aut: the automaton
        :param DFA|NFA U: Universe of permitted words (Sigma^* as default)
        :rtype: bool"""
        t = infixTransducer(aut.Sigma, True)
        a = aut.dup().star()
        return (~(t.runOnNFA(a))).emptyP()

    def satisfiesP(self, aut):
        """Satisfaction of the code property by the automaton language

        :param DFA|NFA aut: the automaton
        :rtype: bool"""
        return self.notSatisfiesW(aut) == (None, None)


class PrefixProp(TrajProp, FixedProp):
    """ Prefix Property

    .. inheritance-diagram:: PrefixProp"""
    def __init__(self, t):
        name = ('Fixed','Prefix_codes')
        IATProp.__init__(self, t, name)

    def satisfiesPrefixP(self, aut):
        """Satisfaction of property by the automaton language: faster than satisfiesP

        :param DFA|NFA aut: the automaton
        :rtype: bool"""
        if isinstance(aut, fa.DFA):
            a = aut.dup()
            a.trim()
            for s in a.Final:
                if a.delta.get(s, {}):
                    return False
            return True
        else:
            return IATProp.satisfiesP(self, aut)


class SuffixProp(TrajProp, FixedProp):
    """ Suffix Property

    .. inheritance-diagram:: SuffixProp"""

    def __init__(self, t):
        name = ('Fixed','Suffix_codes')
        IATProp.__init__(self, t, name)


class InfixProp(PrefixProp, SuffixProp):
    """ Infix Property

    .. inheritance-diagram:: InfixProp"""

    def __init__(self, t):
        name = ('Fixed','Infix_codes')
        IATProp.__init__(self, t, name)


class OutfixProp(PrefixProp, SuffixProp):
    """ Outfix Property

    .. inheritance-diagram:: PrefixProp"""

    def __init__(self, t):
        name = ('Fixed','Outfix_codes')
        IATProp.__init__(self, t, name)


class HypercodeProp(InfixProp, OutfixProp):
    """ Hypercode Property

    .. inheritance-diagram:: PrefixProp"""

    def __init__(self, t):
        name = ('Fixed','Hypercodes')
        IATProp.__init__(self, t, name)


class ErrCorrectProp(ErrDetectProp):
    """ Error Correcting Property

    .. inheritance-diagram:: ErrCorrectProp"""
    def __init__(self, t):
        super(ErrCorrectProp, self).__init__(t)
        self.Name = "Error Correcting Codes"

    def notSatisfiesW(self, aut):
        """Satisfaction of the code property by the automaton language

        :param DFA|NFA aut: the automaton
        :rtype: tuple"""
        return self.Aut.inverse().outIntersection(aut).nonFunctionalW()

    def satisfiesP(self, aut):
        """Satisfaction of the property by the automaton language

        .. seealso:: S. Konstantinidis: Transducers and the Properties of Error-Detection, Error-Correction and
            Finite-Delay Decodability. Journal Of Universal Computer Science 8 (2002), 278-291.

        :param DFA|NFA aut: the automaton
        :rtype: bool"""
        return self.notSatisfiesW(aut) == (None, None, None)

    def notMaximalW(self, aut, U=None):
        """Tests if the language is maximal w.r.t. the property

        :param DFA|NFA aut: the automaton
        :param DFA|NFA U: Universe of permitted words (Sigma^* as default)
        :rtype: bool"""
        return ErrDetectProp(self.Aut.inverse().composition(self.Aut)).notMaximalW(aut, U)


def buildUDCodeProperty(alphabet):
    """Builds a UDCodeProp (from thin air ;-)

    :param set alphabet: alphabet
    :rtype: UDCodeProp"""
    return UDCodeProp(alphabet)


buildUDCodeProp = buildUDCodeProperty


def buildTrajPropS(regex, sigma):
    """Builds a TrajProp from a string regexp

    :param str regex: the regular expression
    :param set sigma: alphabet
    :rtype: TrajProp"""
    au = reex.str2regexp(regex).toNFA()
    return TrajProp(au, sigma)


def buildIATPropF(fname):
    """Builds a IATProp from a FAdo SFT file

    :param str fname: file name
    :rtype: IATProp"""
    t = fio.readOneFromFile(fname)
    return IATProp(t)


def buildIATPropS(s):
    """Builds a IATProp from a FAdo SFT string

    :param str s: string containing SFT
    :rtype: IATProp"""
    t = fio.readOneFromString(s)
    return IATProp(t)


def buildIPTPropF(fname):
    """Builds a IPTProp from a FAdo SFT file

    :param str fname: file name
    :rtype: IPTProp"""
    t = fio.readOneFromFile(fname)
    return IPTProp(t)


def buildIPTPropS(s):
    """Builds a IPTProp from a FAdo SFT string

    :param str s: file name
    :rtype: IPTProp"""
    t = fio.readOneFromString(s)
    return IPTProp(t)


def buildErrorDetectPropF(fname):
    """ Builds an Error Detecting Property

    :param str fname: file name
    :rtype: ErrDetectProp"""
    t = fio.readOneFromFile(fname)
    return IPTProp(t)


def buildErrorDetectPropS(s):
    """ Builds an Error Detecting Property from string

    :param str s: transducer string
    :rtype: ErrDetectProp"""
    t = fio.readOneFromString(s)
    return IPTProp(t)


def buildErrorCorrectPropF(fname):
    """ Builds an Error Correcting Property

    :param str fname: file name
    :rtype: ErrCorrectProp"""
    t = fio.readOneFromFile(fname)
    return ErrCorrectProp(t)


def buildErrorCorrectPropS(s):
    """ Builds an Error Correcting Property from string

    :param str s: transducer string
    :rtype: ErrCorrectProp"""
    t = fio.readOneFromString(s)
    return ErrCorrectProp(t)


def buildPrefixProperty(alphabet):
    """ Builds a Prefix Code Property

    :param set alphabet: alphabet
    :rtype: PrefixProp"""
    return PrefixProp(prefixTransducer(alphabet))


def buildSuffixProperty(alphabet):
    """ Builds a Suffix Code Property

    :param set alphabet: alphabet
    :rtype: PrefixProp"""
    return SuffixProp(suffixTransducer(alphabet))


def buildInfixProperty(alphabet):
    """ Builds a Suffix Code Property

    :param set alphabet: alphabet
    :rtype: PrefixProp"""
    return InfixProp(infixTransducer(alphabet))


def buildOutfixProperty(alphabet):
    """ Builds a Outfix Code Property

    :param set alphabet: alphabet
    :rtype: PrefixProp"""
    return OutfixProp(outfixTransducer(alphabet))


def buildHypercodeProperty(alphabet):
    """ Builds a Hypercode Property

    :param set alphabet: alphabet
    :rtype: PrefixProp"""
    return HypercodeProp(hypercodeTransducer(alphabet))


def createInputAlteringSIDTrans(n, sigmaSet):
    """Create an input-altering SID transducer based

    :param int n: max number of errors
    :param set sigmaSet: alphabet
    :return: a transducer representing the SID channel
    :rtype: SFT"""
    new = SFT()
    new.setSigma(sigmaSet)
    new.setOutput(sigmaSet)
    init = new.stateIndex((0, None), True)
    new.addInitial(init)
    for sy in new.Sigma:
        new.addTransition(init, sy, sy, init)
    i = 1
    while i <= n:
        d1 = new.stateIndex((i, None), True)
        new.addFinal(d1)
        for s in new.Sigma:
            new.addTransition(d1, s, s, d1)
            d = new.stateIndex((i, s), True)
            new.addFinal(d)
            if i == 1:
                new.addTransition(init, s, Epsilon, d)
                for s1 in new.Sigma - {s}:
                    new.addTransition(init, s, s1, d1)
            else:
                bar = new.stateIndex((i - 1, None))
                new.addTransition(bar, s, Epsilon, d1)
                new.addTransition(bar, Epsilon, s, d1)
                for s1 in new.Sigma:
                    foo = new.stateIndex((i - 1, s))
                    new.addTransition(foo, s1, Epsilon, d)
                    if s1 != s:
                        new.addTransition(bar, s, s1, d1)
                        new.addTransition(foo, Epsilon, s1, d1)
                        for s2 in new.Sigma:
                            new.addTransition(foo, s2, s1, d1)
            for s1 in new.Sigma - {s}:
                new.addTransition(d, s1, s1, d1)
        i += 1
    return new


def editDistanceW(auto):
    """Compute the edit distance of a given regular language accepted by the NFA via Input-altering transducer.

    .. attention::
       language should have at least two words

    .. seealso:: Lila Kari, Stavros Konstantinidis, Steffen Kopecki, Meng Yang. An efficient algorithm for computing
        the edit distance of a regular language via input-altering transducers. arXiv:1406.1041 [cs.FL]

    :param NFA auto: language recogniser
    :return: The edit distance of the given regular language plus a witness pair
    :rtype: tuple"""

    def _normalize(tr):
        tr.States = [(s1, s2, s3, s4) for (((s1, s2), s3), s4) in tr.States]

    def _addTNone(tr, au, n1, s1, si, so, symb1, symb2, l):
        for i1 in au.delta[si].get(symb1, []):
            for i2 in au.delta[so].get(symb2, []):
                new = tr.stateIndex((n1, None, i1, i2), True)
                l.add(new)
                if i1 in au.Final and i2 in au.Final:
                    tr.addFinal(new)
                tr.addTransition(s1, symb1, symb2, new)

    def _addT(tr, au, n1, s1, aa, si, so, symb1, symb2, l):
        for i1 in au.delta[si].get(symb1, []):
            for i2 in au.delta[so].get(symb2, []):
                new = tr.stateIndex((n1, aa, i1, i2), True)
                l.add(new)
                if i1 in au.Final and i2 in au.Final:
                    tr.addFinal(new)
                tr.addTransition(s1, symb1, symb2, new)

    def _noneLoops(tr, au, l):
        todo = copy.deepcopy(l)
        done = set()
        while todo:
            s = todo.pop()
            done.add(s)
            (a, _, s3, s4) = tr.States[s]
            for b in tr.Sigma:
                for i1 in au.delta[s3].get(b, []):
                    for i2 in au.delta[s4].get(b, []):
                        new = tr.stateIndex((a, None, i1, i2), True)
                        if i1 in au.Final and i2 in au.Final:
                            tr.addFinal(new)
                        if new not in done:
                            l.add(new)
                            todo.add(new)
                        tr.addTransition(s, b, b, new)

    def extendTrans(tr, au, lln, ll):
        """ Extend the transducer

        :param set ll: last level added states (not None)
        :param set lln: last level added states (None)
        :param SFT tr: the old transducer to be updated
        :param NFA au: the automaton recogize the regular language
        :return: last level added states lln, ll
        :rtype: tuple

        .. attention::
           NFA au must have its state names as indexes (renameStates())"""
        newll, newlln = set(), set()
        for s in lln:
            (m, _, s3, s4) = tr.States[s]
            for a in tr.Sigma:
                _addTNone(tr, au, m + 1, s, s3, s4, a, Epsilon, newlln)
                _addTNone(tr, au, m + 1, s, s3, s4, Epsilon, a, newlln)
                for b in tr.Sigma - {a}:
                    _addTNone(tr, au, m + 1, s, s3, s4, a, b, newlln)
        for s in ll:
            (m, a, s3, s4) = tr.States[s]
            for b in tr.Sigma:
                _addT(tr, au, m + 1, s, a, s3, s4, b, Epsilon, newll)
                if b != a:
                    _addTNone(tr, au, m + 1, s, s3, s4, Epsilon, b, newlln)
                    for c in tr.Sigma - {b}:
                        _addTNone(tr, au, m + 1, s, s3, s4, c, b, newlln)
        for s in newll:
            (m, a, s3, s4) = tr.States[s]
            for b in tr.Sigma - {a}:
                _addTNone(tr, au, m, s, s3, s4, b, b, newlln)
        _noneLoops(tr, au, newlln)
        return newlln, newll

    auto = auto.toNFA().renameStates().addEpsilonLoops()
    #Todo: avoid use of inIntersectionSlow
    t = createInputAlteringSIDTrans(1, auto.Sigma).inIntersectionSlow(auto).outIntersection(auto)
    _normalize(t)
    lvn, lv = set(), set()
    for i in range(len(t.States)):
        n0, a0 = t.States[i][:2]
        if n0 != 0:
            if a0 is None:
                lvn.add(i)
            else:
                lv.add(i)
    errn = 1
    while True:
        w = t.nonEmptyW()
        if w != (None, None):
            return errn, w
        lvn, lv = extendTrans(t, auto, lvn, lv)
        errn += 1


def exponentialDensityP(aut):
    """ Checks if language density is exponential

    Using breadth first search (BFS)

    .. attention::
        aut should not have Epsilon transitions

    :param NFA aut: the representation of the language
    :rtype: bool"""
    import fractions
    from collections import deque

    def _listgcd(ll):
        if len(ll) == 1:
            return ll[0]
        return fractions.gcd(ll[0], _listgcd(ll[1:]))

    components = aut.trim().stronglyConnectedComponents()
    for c in components:
        if not len(c) == 1:
            g = aut.dup()
            g.deleteStates(list(set(g.indexList(g.States)) - set(c)))
            marked, levelNum, levelSym = [], [], []
            for _ in g.States:
                marked.append(False)
                levelNum.append([])
                levelSym.append([])
            v = g.States[0]
            queue = deque()
            queue.append(v)
            marked[g.stateIndex(v)] = True
            levelNum[g.stateIndex(v)].append(0)
            while len(queue):
                t = queue.popleft()
                symbols = [s for s in g.delta.get(g.stateIndex(t), {})]
                edges = [g.delta[g.stateIndex(t)][k] for k in symbols]
                for s in symbols:
                    if s not in levelSym[levelNum[g.stateIndex(t)][0]]:
                        levelSym[levelNum[g.stateIndex(t)][0]].append(s)
                for e in edges:
                    for o in e:
                        levelNum[o].append(levelNum[g.stateIndex(t)][0] + 1)
                        if not marked[o]:
                            marked[o] = True
                            queue.append(g.States[o])
            if [l for l in levelSym if len(l) > 1]:
                return True
            lengths = []
            for s in levelNum:
                first = s[0]
                for n in [m for m in s if m - first]:
                    lengths.append(n - first)
            d = _listgcd(lengths)
            if d == 1:
                return True
            pattern = ""
            for i in range(0, d):
                pattern = pattern + levelSym[i][0]
            for j in range(d, max(lengths)):
                if not levelSym[j][0] == pattern[j % d]:
                    return True
    return False


def long2base(n, q):
    """ Maps n to a list of digits corresponding to the base q representation of n in reverse order

    :param int n: a positive integer
    :param int q: base to represent n
    :return: list of q-ary 'digits', that is, elements of {0,1,..,q-1}
    :rtype: list"""
    lt = []
    while n > 0:
        lt.append(n % q)
        n /= q
    return lt


def list2string(lt, dy):
    """
    :param list lt: list of nonnegative integers from some set {0,1,...,q-1}
    :param dict dy: mapping from {0,1,...,q-1} to some alphabet symbols
    :return: string of symbols corresponding to the integers in lt
    :rtype: str"""
    return "".join([dy[i] for i in lt])


def notUniversalStatW(a, l, maxIter=20000):
    """ Tests statistically whether the NFA a is l-non-universal, by evaluating a on <= maxIter randomly chosen words of length l

    :type a: NFA
    :param int l: nonnegative integer
    :type maxIter: int
    :return: (w,i) where w is the word found at i-th try; or (None, i) after i tries
    :rtype: tuple"""
    if l == 0:  # first handle the case of the empty word
        if a.evalWordP(Epsilon):
            return None, 1
        else:
            return Epsilon, 1
    lt = [s for s in a.Sigma]
    # dictionary dy is used to refer to each alphabet symbol in lt using a numeric index
    dy = {i: lt[i] for i in range(0, len(lt))}
    q = len(lt)
    numWords = q ** l
    tmp = set()  # will remember in tmp the distinct words generated so far
    random.seed()
    for j in range(0, maxIter):
        # next pick one word out of the possible numWords of length l; specifically pick
        # the i-th one, where i is randomly chosen. For this, choose a random int n whose
        # base q representation is of the form '1d_1...d_l', where the l d's are digits
        # in {0,1,...,q-1}, so the desired number i is the one represented as 'd_1...d_l'
        n = random.randint(numWords, 2 * numWords - 1)
        w = list2string(long2base(n, q), dy)[:l]
        if not a.evalWordP(w):
            return w, j + 1
        tmp.add(w)  # Note: the current random word could be equal to a past one !!
        if len(tmp) == numWords:
            return None, numWords
    return None, maxIter


def symmAndRefl(t, ipt=False):
    """ Return the  transducer      t | t.inverse,  if ipt is True;
        return the transducer  t | t.inverse | id,  otherwise

    :type t: SFT
    :type ipt: bool
    :rtype: SFT"""
    t1 = t | t.inverse()
    if ipt is True:
        return t1
    t2 = SFT()
    s0 = t2.addState()
    t2.addInitial(s0)
    t2.addFinal(s0)
    for sym in t.Sigma:
        t2.addTransition(s0, sym, sym, s0)
    return t1 | t2


def constructCode(n, l, p, ipt=False, seed=None):
    """ Returns up to n words of length l satisfying the property p, the first one being seed.
        If ipt is True, the property is assumed to be input-preserving transducer type

    :type p: IATProp|IPTProp
    :type l: int
    :type n: int
    :type seed: str
    :type ipt: bool
    :rtype: list"""
    if n == 0:
        return []
    if l == 0:
        return [Epsilon]
    t = symmAndRefl(p.Aut, ipt)  # make a version of p.Aut that is symmetric and reflexive
    if seed is None or len(seed) != l:
        a0 = list(t.Sigma)[0]
        seed = "".join([a0 for _ in range(l)])
    lt = [seed]
    cnt, more = 1, True
    while more and cnt < n:
        a = fl.FL(lt).trieFA().toNFA()
        b = t.runOnNFA(a)
        (w, i) = notUniversalStatW(b, l)
        if w is None:
            more = False
        else:
            cnt += 1
            lt.append(w)
    return lt



# -*- coding: utf-8 -*-
""" General Boltzmann Generator environment

@author: Rogério Reis & Nelma Moreira

This is part of U{FAdo project <http://www.ncc.up.pt/FAdo>}.

@version: 0.9.5

Regular expression classes and manipulation

@copyright: 1999-2011 Rogério Reis & Nelma Moreira {rvr,nam}@dcc.fc.up.pt

B{Naming convention:} methods suffixed by P have boolean return.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA."""

import random
import math

from FAdo.reex import *


class emptyword(epsilon):
    def generating(self, x):
        return x

    def generating_s(self, x):
        return str(x)

    def sampler(self, x):
        """ None if empty"""
        return [epsilon()]


class singleton(regexp):
    def generating(self, x):
        return x

    def generating_s(self, x):
        return str(x)

    def sampler(self, x):
        """ None if empty"""
        return [self.val]


class union(disj):
    def __init__(self, *args):
        self.gen = {}
        disj.__init__(self, *args)

    def generating(self, x):
        return self.gen.setdefault(x, self.arg1.generating(x) + self.arg2.generating(x))

    def generating_s(self, x):
        return "(" + str(self.arg1.generating_s(x)) + "+" + str(self.arg2.generating_s(x)) + ")"

    def sampler(self, x):
        pa = self.arg1.generating(x) * 1.0 / (self.arg1.generating(x) + self.arg2.generating(x))
        if rnd_bernoulli(pa):
            return self.arg1.sampler(x)
        else:
            return self.arg2.sampler(x)


class prod(concat):
    def __init__(self, *args):
        self.gen = {}
        concat.__init__(self, *args)

    def generating(self, x):
        return self.gen.setdefault(x, self.arg1.generating(x) * self.arg2.generating(x))

    def generating_s(self, x):
        return str(self.arg1.generating_s(x)) + "*" + str(self.arg2.generating_s(x))

    def sampler(self, x):
        return self.arg1.sampler(x) + self.arg2.sampler(x)

    def string(self):
        return "(%s.%s)" % (self.arg1, self.arg2)


class seq(star):
    def __init__(self, *args):
        self.gen = {}
        star.__init__(self, *args)

    def generating(self, x):
        return self.gen.setdefault(x, 1.0 / (1 - self.arg.generating_s(x)))

    def generating_s(self, x):
        return str(1) + "/(" + str(1) + "-" + str(self.arg.generating_s(x)) + ")"

    def sampler1(self, x):
        if rnd_bernoulli(self.arg.generating(x)):
            a = self.sampler1(x)
            if a:
                return [self.arg.sampler(x)] + a
            else:
                return [self.arg.sampler(x)]
        else:
            return epsilon()

    def sampler(self, x):
        k = rnd_geometric(self.arg.generating(x))
        return [self.arg.sampler(x) for i in xrange(k + 1)]


    def string(self):
        return "(%s)*" % self.arg


def rnd_bernoulli(p):
    """
  p(0) = 1 - p; p(1) = p
  """
    u = random.random()
    if u <= p:
        return 1
    else:
        return 0


def rnd_bernoulli_l(l):
    """
  sum(l)=1

  """
    u = random.random()
    k = len(l)
    p = 0
    for i in range(k):
        p += l[i][0]
        if u <= p:
            return l[i][1](i)


def rnd_geometric(p):
    """
  p(k) =  p(1-p)^(k-1)
  """
    if p == 1 or p == 0:
        raise Exception
    u = random.random()
    pk = 1 - p
    s = 0
    k = 0
    while u > s:
        s += pk
        pk *= p
        k += 1
    return k


def rnd_geometric0(p):
    """
  p(k) =  p(1-p)^(k-1)

  """

    u = random.random()
    if p == 1:
        return 1
    else:
        return int(math.log(u) / math.log(1 - p)) + 1


def rnd_poisson(l, t=0):
    """
  p(k) = {\mu^k \over k!} \exp(-\mu)
  """
    u = random.random()
    p = math.exp(-l)
    k = 0
    s = 0
    while u < s:
        s += p
        k += 1
        p = 1.0 * l * p / (k + 1)
    return k


def reject(g, x, n, e=0):
    while 1:
        a = g.sampler(x)
        if n * (1 - e) <= len(a) <= n * (1 + e):
            return a


def qui(k):
    return k * (1 - 1 / (math.exp(k)) + (k / math.exp(2 * k)))



# -*- coding: utf-8 -*-
"""**Finite two-way automata manipulation.**

Deterministic and non-deterministic two-way automata manipulation, conversion and evaluation.

.. *Authors:* Rogério Reis & Nelma Moreira

.. *This is part of FAdo project*   http://fado.dcc.fc.up.pt.

.. *Copyright:* 2014 Rogério Reis & Nelma Moreira {rvr,nam}@dcc.fc.up.pt

.. This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   675 Mass Ave, Cambridge, MA 02139, USA."""
import fa
from abc import abstractmethod
from common import *


class WordEval(object):
    """Class for word evaluation configuration
    :type Aut: TFA"""
    def __str__(self):
        return str(self.Pos), str(self.Word), str(self.Aut), str(self.State)

    def __repr__(self):
        return self.__str__()

    def __init__(self, word, sti):
        self.Word = word
        self.RLi = len(self.Word) - 1
        self.Pos = None
        self.Trace = None
        self.Aut = None
        self.State = sti

    def _start(self, aut):
        """Initialises evaluation configuration

        :param aut: automaton"""
        self.Aut = aut
        self.State = self.Aut.Initial
        self.Pos = 0
        self.Trace = []

    def eval1(self):
        """Eval one step

        :raises TAFSignal: accordingly to execution"""
        if self.Pos == -1:
            sym = TFA.LMark
        elif self.Pos > self.RLi:
            sym = TFA.RMark
        else:
            sym = self.Word[self.Pos]
        (st, mv) = self.Aut.evalSymbol(self.State, sym)
        if (self.State, self.Pos, mv) in self.Trace:
            raise TFARejectLoop
        else:
            self.Trace.append((self.State, self.Pos, mv))
        if self.Pos > self.RLi:
            if self.Aut.finalP(st):
                raise TFAAccept
            else:
                raise TFARejectNonFinal
        self.State = st
        self.Pos += mv

    def eval(self, aut):
        """Evaluate the word

        :param aut: automaton
        :type aut: TFA

        :raises TFASignal: accordinliy to the evaluation"""
        self._start(aut)
        while True:
            self.eval1()


class TFA(fa.FA):
    """ Base class for two-way automata"""
    LMark = "@"
    RMark = "#"

    @abstractmethod
    def dotFormat(self, size):
        raise NotImplementedError

    @abstractmethod
    def succintTransitions(self):
        raise NotImplementedError

    @abstractmethod
    def dotDrawTransition(self, st1, sym, st2, sep):
        raise NotImplementedError

    def __ne__(self, other):
        raise NotImplementedError

    def __eq__(self, other):
        raise NotImplementedError

    @abstractmethod
    def evalSymbol(self):
        raise NotImplementedError

    Dir = ["R", "L"]

    @abstractmethod
    def addTransition(self):
        """Add Transition"""
        raise NotImplementedError


class TDFA(TFA):
    """ Class of Deterministic Two-Way Finite Automata"""

    @staticmethod
    def trD(mv):
        """ Translate mv from int to str

        :param mv: move
        :type mv: int
        :rtype: str"""
        if mv == 1:
            return "R"
        else:
            return "L"

    def addTransition(self, st1, sym, st2, mov):
        """Add transition to DFTA

        :param st1: starting state
        :param st2: ending state
        :param sym: symbol consumed (can be a border marker)
        :param mov: direction of the next move
        :type st1: int
        :type st2: inst"""
        if st1 not in self.delta:
            self.delta[st1] = dict()
            self.delta[st1][sym] = (st2, mov)
        else:
            self.delta[st1][sym] = (st2, mov)

    def succintTransitions(self):
        """ Collects the transition information in a compact way suitable for graphical representation.
        :rtype: list of tupples

        .. note:
            tupples in the list are stateout, label, statein

        .. versionadded: 0.9.8"""
        foo = dict()
        for s in self.delta:
            for c in self.delta[s]:
                k = (s, self.delta[s][c][0])
                if k not in foo:
                    foo[k] = []
                foo[k].append(c, self.delta[s][c][1])
        l = []
        for k in foo:
            cs = foo[k]
            s = "%s,%s" % (str(cs[0][0]), self.trD(cs[0][1]))
            for c in cs[1:]:
                s += "; %s,%s" % (str(c[0]), self.trD(c[1]))
            l.append((str(self.States[k[0]]), s, str(self.States[k[1]])))
        return l

    def dotDrawTransition(self, st1, sym, st2, sep):
        pass

    def evalSymbol(self, init, sym):
        """Eval a symbol from a state

        :param init: starting state
        :type init: int
        :param sym: symbol read
        :returns: tupple (state,movement)
        :rtype: tuple
        :raises TFARejectBlocked: if transition is not valid"""
        if init not in self.delta or sym not in self.delta[init]:
            raise TFARejectBlocked
        return self.delta[init][sym]
